#include "mainwindow.h"
#include "mesh.h"
#include "ui_mainwindow.h"
#include <QtDebug>
#include <QHash>


MainWindow::~MainWindow()
{
    delete ui;
    delete mesh;
}

MainWindow::MainWindow(QWidget *parent)
    : QMainWindow(parent), ui(new Ui::MainWindow)
{
    ui->setupUi(this);

    getUiInputWidgets();

    // find max number of neighbours for any node:
    int maxNbors = findMaxNbors();

    // creating the mesh
    mesh = new SimulationMesh(tempDSpins.size(), distDSpins.size(), maxNbors, paramsDSpins.size());

    int tmpFrom, tmpTo;

    // filling the neighbour-matrix
    for (int i = 0; i < distDSpins.size(); i++)
    {        
        tmpFrom = getLinkfrom(i);
        tmpTo = getLinkto(i);
        if(tmpFrom < 0 || tmpTo < 0 || tmpFrom == tmpTo)
        {
            qDebug() << "Error: one of the links had an invalid from- or to-property";
        }
        mesh->addLink(tmpFrom,tmpTo);
    }

    editMeshButton = ui->editButton;
    stepButton = ui->stepButton;
    playButton = ui->playButton;

    speedBox = ui->speedDspin;

    connect(stepButton, SIGNAL(clicked()), this, SLOT(stepPressed()));
    connect(playButton, SIGNAL(clicked()), this, SLOT(playPressed()));
    connect(editMeshButton, SIGNAL(clicked()), this, SLOT(editMeshPressed()));

    simRunning = false;

    stepSimTimer = new QTimer(this); // making the timer a child of "this" ensures it is automatically destroyed along with "this".

    connect(stepSimTimer, SIGNAL(timeout()), this, SLOT(simTimedout()));

    editingMesh = true;
    editMeshPressed();

}

void MainWindow::updateTempColors()
{
    for (int i = 0; i < tempDSpins.size(); i++)
    {
        bool ign = mesh->getNodeIgn(i);
        QColor color = Qt::green;
        if (ign)
            color = Qt::red;
        QPalette pal = QPalette();
        pal.setColor(QPalette::Window, color);
        tempDSpins[i]->setPalette(pal);
    }


}
void MainWindow::updateFuelColors()
{
    for (int i = 0; i < fuelDSpins.size(); i++)
    {
        float f = mesh->getNodeFuel(i);
        QColor color = Qt::green;
        if (f <= 0.0)
            color = Qt::red;
        QPalette pal = QPalette();
        pal.setColor(QPalette::Window, color);
        fuelDSpins[i]->setPalette(pal);

    }
}

void MainWindow::updateDistColors()// example Qt::green
{
    for (int i = 0; i < distDSpins.size(); i ++)
    {
        float dist = mesh->getNborDist( getLinkfrom(i), getLinkto(i) );
        QColor color = Qt::green;
        if(dist <= 0.0)
            color = Qt::red;
        QPalette pal = QPalette();
        pal.setColor(QPalette::Window, color);
        distDSpins[i]->setPalette(pal);
    }

}

void MainWindow::lockMesh()
{
    for (int i = 0; i < distDSpins.size(); i++)
    {
        distDSpins[i]->setFrame(false);
        distDSpins[i]->setReadOnly(true);
    }
    for (int i = 0; i < tempDSpins.size(); i++)
    {
        tempDSpins[i]->setFrame(false);
        fuelDSpins[i]->setFrame(false);
        tempDSpins[i]->setReadOnly(true);
        fuelDSpins[i]->setReadOnly(true);
    }
}

void MainWindow::unlockMesh()
{
    for (int i = 0; i < distDSpins.size(); i++)
    {
        distDSpins[i]->setFrame(true);
        QPalette pal = QPalette();
        pal.setColor(QPalette::Window, Qt::gray);
        distDSpins[i]->setPalette(pal);
        distDSpins[i]->setReadOnly(false);
    }
    for (int i = 0; i < tempDSpins.size(); i++)
    {
        tempDSpins[i]->setFrame(true);
        fuelDSpins[i]->setFrame(true);
        QPalette palt = QPalette();
        QPalette palf = QPalette();
        palt.setColor(QPalette::Window, Qt::gray);
        palf.setColor(QPalette::Window, Qt::gray);
        tempDSpins[i]->setPalette(palt);
        fuelDSpins[i]->setPalette(palf);
        tempDSpins[i]->setReadOnly(false);
        fuelDSpins[i]->setReadOnly(false);
    }

}

void MainWindow::getUiInputWidgets()
{
    QVariant testVariant;
    QDoubleSpinBox* tempSbp;
    bool ok;
    int tmpIndex;

    // collect parameter dspins
    for(int i = 0; i < ui->paramsPage->children().size(); i++)
    {
        testVariant = ui->paramsPage->children().at(i)->property("paramname");
        if(testVariant.isValid() && !testVariant.toString().isEmpty() && \
           (tempSbp = qobject_cast<QDoubleSpinBox*>(ui->paramsPage->children()[i]))\
           )
        {
            paramsDSpins.append(tempSbp);
        }

    }

    // filling the xxDSpins-arrays with QDoubleSpinBoxes from the meshContentWidget
    for (int i = 0; i < ui->meshContentWidget->children().size(); i++)
    {
        testVariant = ui->meshContentWidget->children().at(i)->property("type");

        if(testVariant.isValid() && !testVariant.toString().isEmpty() \
           && (tempSbp = qobject_cast<QDoubleSpinBox*>(ui->meshContentWidget->children()[i]))\
           )
        {
            QString type = testVariant.toString();

            testVariant = ui->meshContentWidget->children().at(i)->property("nodeindex");
            tmpIndex = testVariant.toInt(&ok);
            if(type == "nodetemp")
            {
                if(ok)
                {
                    tempDSpins.insert(tmpIndex, tempSbp);
                }else{
                    qDebug() << "Error: one of the nodetemp widgets' index was invalid";// << std::endl;
                }

            }else if(type == "nodefuel")
            {
                if(ok)
                {
                    fuelDSpins.insert(tmpIndex, tempSbp);
                }else{
                    qDebug() << "Error: one of the nodefuel widgets' index was invalid";
                }

            }else if(type == "linklength")
            {
                distDSpins.append(tempSbp);

            }else
            {
                qDebug() << "Error: one of the widgets in the mesh had a type-string not recognized: '" << type << \
                        "', or is not a QDoubleSpinBox.";
            }
           tempSbp->setAutoFillBackground(true); // set auto-fill background for all mesh dspins.

        }else
        {
            qDebug() << "Error: of the widgets had invalid or empty type-string.";
        }
    }

    //Error-check:
    if(tempDSpins.size() != fuelDSpins.size())
    {
        qDebug() << "Error: one of the nodes lack either a temperature widget or a fuel widget.";
    }
}

void MainWindow::editMeshPressed()
{
    if(editingMesh)
    {
        editingMesh = false;
        editMeshButton->setText("Edit Mesh Values");
        lockMesh();
        saveMeshValues();
        updateDistColors();
        updateFuelColors();
        updateTempColors();
    }else{
        editingMesh = true;
        editMeshButton->setText("Save Mesh Values");
        unlockMesh();
    }


}

void MainWindow::saveMeshValues()
{

    int from,to;
    // node data:
    for (int i = 0; i < tempDSpins.size(); i++)
    {
        mesh->updateNodeValues(i, tempDSpins[i]->value(), fuelDSpins[i]->value());
    }
    // link-lengths
    for (int i = 0; i < distDSpins.size(); i++)
    {
        from = getLinkfrom(i);
        to = getLinkto(i);
        if(from >= 0 && to >= 0 && from != to)
            mesh->updateDist(from, to, distDSpins[i]->value());
        else
            qDebug() << "ERROR: a link's from or to property was erronous.";

    }
}

// count number of neighbours of each node
// to find max number of neighbours for any node:
int MainWindow::findMaxNbors()
{
	   
    int maxNbors = 0;
    int* nborCount = new int[tempDSpins.size()];
    memset(nborCount, 0, tempDSpins.size()*sizeof(int));
    int tmpFrom, tmpTo;
    for(int i = 0; i < distDSpins.size(); i++)
    {
        tmpFrom = getLinkfrom(i);
        tmpTo = getLinkto(i);
        if(tmpFrom < 0 || tmpTo < 0 || tmpFrom == tmpTo)
        {
            qDebug() << "Error: one of the links had an invalid from- or to-property";
        }
        nborCount[tmpFrom]++;
        if(nborCount[tmpFrom]>maxNbors)
            maxNbors = nborCount[tmpFrom];
        nborCount[tmpTo]++;
        if(nborCount[tmpTo]>maxNbors)
            maxNbors = nborCount[tmpTo];
    }
    delete[] nborCount;
    return maxNbors;

}




/**
  * Helper function
  * returns: from-index, or -1 if error
  */
int MainWindow::getLinkfrom(int spinIndex)
{
    return getDistdspinIntProperty(spinIndex, "linkfrom");
}

/**
  * Helper function
  * returns: to-index, or -1 if error
  */
int MainWindow::getLinkto(int spinIndex)
{
    return getDistdspinIntProperty(spinIndex, "linkto");
}

/**
  * Helper function
  * returns: integer property, or -1 if error
  */
int MainWindow::getDistdspinIntProperty(int spinIndex, const char* propertyname)
{
    bool ok;
    int tmpValue;
    QVariant testVariant;
    testVariant = distDSpins[spinIndex]->property(propertyname);
    tmpValue = testVariant.toInt(&ok);
    if(!ok)
    {
        qDebug() << "Error: one of the links had an invalid property '" << propertyname << "'.";
        return -1;
    }
    return tmpValue;

}

void MainWindow::stepSimulation()
{
    // run simulation/ calculations
    mesh->simulate();

    // write back new values
    for (int i = 0; i < tempDSpins.size(); i++)
    {
        tempDSpins[i]->setValue(mesh->getNodeTemp(i));
        fuelDSpins[i]->setValue(mesh->getNodeFuel(i));
        updateDistColors();
        updateFuelColors();
        updateTempColors();
    }
}

void MainWindow::prepareSimulation()
{
    QString name;

    // parameter values:
    for (int i = 0; i < paramsDSpins.size(); i++)
    {
        name = paramsDSpins[i]->property("paramname").toString();
        if (!name.isEmpty())
            mesh->updateParameter(name, paramsDSpins.at(i)->value());
        else
            qDebug() << "ERROR: a parameterwidget's name-property was erronous.";
    }


}


// Disable all inputs while running the simulation
void MainWindow::lockOptions(bool readOnly)
{
	QHash<int,QDoubleSpinBox*>::iterator i;
	for (i = tempDSpins.begin(); i != tempDSpins.end(); ++i)
		(i.value())->setReadOnly(readOnly);
	for (i = fuelDSpins.begin(); i != fuelDSpins.end(); ++i)
		(i.value())->setReadOnly(readOnly);
	QList<QDoubleSpinBox*>::iterator j;
	for (j = distDSpins.begin(); j != distDSpins.end(); ++j)
		(*j)->setReadOnly(readOnly);
	for (j = paramsDSpins.begin(); j != paramsDSpins.end(); ++j)
		(*j)->setReadOnly(readOnly);
	speedBox->setReadOnly(readOnly);

}


void MainWindow::playPressed()
{
	double speed = speedBox->value();
	if(speed <= 0.000001)
		speed = 0.000001;

	prepareSimulation();

	if(simRunning)
	{
		playButton->setText(">");
		simRunning = false;
		stepSimTimer->stop();
		lockOptions(false);
	}else{
		playButton->setText("||");
		simRunning = true;
		lockOptions(true);
		stepSimTimer->start(1000/speed);
	}

}


void MainWindow::stepPressed()
{
	prepareSimulation();
	stepSimulation();    

}

void MainWindow::simTimedout()
{
	stepSimulation();    
}


