#include "mesh.h"
#include <QtDebug>

#define NBORS(row,col) nborMatrix[((row)*(maxNbors))+(col)]
#define DISTS(row,col) nborDists[((row)*(maxNbors))+(col)]
SimulationMesh::SimulationMesh(int numNodes, int numLinks, int maxNbors, int numParams)
{
    this->numNodes = numNodes;
    this->numLinks = numLinks;
    this->maxNbors = maxNbors;
    this->numParams = numParams;
    numNbors = new int[numNodes];
    memset(numNbors, 0, numNodes*sizeof(int));
    nodeTemps = new double[numNodes];
    nodeFuels = new double[numNodes];
    nborMatrix = new int[numNodes*maxNbors];//nborMatrix has neighbour-count in first row
    // fill nborMatrix with non-valid numbers as "empty"-indicators:
    // not necessary?
    //for(int i = 0; i < numNodes*(maxNbors+1); i++)
    //    nborMatrix[i] = -1;
    // instead, set the matrix to zero, indicating no node has filled in neigbhours yet:
    memset(nborMatrix, 0, numNodes*maxNbors*sizeof(int));
    nborDists = new double[numNodes*maxNbors];
    memset(nborDists, 0, numNodes*maxNbors*sizeof(double));
    nodeIgn = new bool[numNodes];
    memset(nodeIgn,0,numNodes*sizeof(bool));
    nodeIgn[5] = true; //ignite one tilfeldig node

}
SimulationMesh::~SimulationMesh()
{
    delete[] numNbors;
    delete[] nodeTemps;
    delete[] nodeFuels;
    delete[] nborDists;
    delete[] nborMatrix;
    delete[] nodeIgn;
}

/**
  * NB: If no parameter named 'name' exist in the hash, a new hash-post is made
  */
void SimulationMesh::updateParameter(QString name, double value)
{
    parameters[name] = value;
    if(parameters.size() > numParams)
    {
        qDebug() << "ERROR: The number of parameters has exceeded the foretold amount.";
    }

}

void SimulationMesh::updateNodeValues(int nodeidx, double temp, double fuel)
{
    nodeTemps[nodeidx] = temp;
    nodeFuels[nodeidx] = fuel;

}

void SimulationMesh::simulate()
{
    // get parameters:
    int thisNumNbors;
    float surroundTemp;
    float surroundAirTemp;
    float deltaT = parameters.value("deltat");
    float igniteTemp = parameters.value("ignitetemp");
    float fireTemp = parameters.value("firetemp");
    float d = 0.0;
    float tempScaling = parameters.value("tempscaling");
    float fuelScaling = parameters.value("fuelscaling");
    float airTemp = parameters.value("airtemp");
    float spreadSpeed = 10.0; //parameters.value("spreadSpeed");

    for (int tid = 0; tid < numNodes; tid++)
    {

        thisNumNbors = numNbors[tid];
        if (nodeIgn[tid]) // if this node has ignited
        {
            for (int i = 0; i < thisNumNbors; i++) // for each link
            {
                if (DISTS(tid,i) > 0.0) // if link is not already burned
                {
                    DISTS(tid,i) -= spreadSpeed * deltaT;
                    // if the link now has burned down, ignite the other end
                    if (DISTS(tid,i) <= 0.0)
                    {
                        nodeIgn[NBORS(tid,i)] = true;

                    }
                }

            }
        }
        float heatProduced = 0.0;
        float heatLost = 0.2 * (nodeTemps[tid]-airTemp);

        // node is burning
        if (nodeIgn[tid] && nodeFuels[tid] > 0.0)
        {
            heatProduced = 0.1 * nodeTemps[tid];
            nodeFuels[tid] -= (deltaT * fuelScaling);
        }
        nodeTemps[tid] += heatProduced-heatLost;


    }



}


/**
  * TODO: Make this a QtScript
  */
void SimulationMesh::simulate_old()
{
    // get parameters:
    int thisNumNbors;
    float surroundTemp;
    float surroundAirTemp;
    float deltaT = parameters.value("deltat");
    float igniteTemp = parameters.value("ignitetemp");
    float fireTemp = parameters.value("firetemp");
    float d = 0.0;
    float* tmpNodeTemp = new float[numNodes];
    float tempScaling = parameters.value("tempscaling");
    float fuelScaling = parameters.value("fuelscaling");
    float airTemp = parameters.value("airtemp");

    for (int tid = 0; tid < numNodes; tid++)
    {
        surroundTemp = 0.0;
        thisNumNbors = numNbors[tid];
        for (int i = 0; i < thisNumNbors; i++)
        {
            d = DISTS(tid,i);
            surroundTemp += (nodeTemps[NBORS(tid,i)] - nodeTemps[tid]) / ((1.0 + d) * (1.0 + d));
        }
        surroundTemp *= tempScaling * deltaT;
        tmpNodeTemp[tid] = surroundTemp + nodeTemps[tid]; // add the previous temp.
        if (
                ((igniteTemp - tmpNodeTemp[tid]) < 2.0) &&
                (nodeFuels[tid] > -1.0)
                )
        {
            surroundAirTemp = (fireTemp - nodeTemps[tid]);
            //tmpNodeTemp[tid] = fireTemp + surroundTemp;
            nodeFuels[tid] -= (deltaT * fuelScaling);
        }else{
            surroundAirTemp = (airTemp - nodeTemps[tid]);
            // loss to air temperature and similar here.
        }
        tmpNodeTemp[tid] = nodeTemps[tid] + ((surroundTemp + surroundAirTemp) * tempScaling * deltaT);
    }

    for (int i = 0; i < numNodes; i++)
    {
        nodeTemps[i] = tmpNodeTemp[i];
    }

    delete[] tmpNodeTemp;

}

/** Create a link between to nodes by adding them to each other's neigbhourlist.
  * returns: zero on success, index of (first) found neighbour
  * if neighbour is already in one of the lists.
  * -1 if trying to go beyond maxNbors.
  */
int SimulationMesh::addLink(int from, int to)
{
    // Create the first half of the link
    int firsterr = addHalfLink(from,to);
    if(firsterr)
        return firsterr;
    // If sucess so far, create second half of the link
    return addHalfLink(to,from);

}

/** Create one half of a link by adding the "to" node to "from"'s neighbourlist.
  * returns: zero on success, index of (first) found neighbour
  * if neighbour is already in one of the lists.
  * -1 if trying to go beyond maxNbors.
  */
int SimulationMesh::addHalfLink(int from, int to)
{
    int nborsCount = numNbors[from];
    if(nborsCount >= maxNbors)
    {
        qDebug() << "ERROR: Trying to fill in more than maxNbors for node " << from << ".";
        return -1;
    }

    //check that "to" is not already in "from"'s neighbourlist
    for(int i = 0; i < nborsCount; i++)
    {
        if(NBORS(from,i) == to)
        {
            qDebug() << "ERROR: Node " << to << " already is in node " << \
                    from << "'s neighbour-list (at column " << i << ").";
            return i;
        }
    }
    // all is well so far - insert new neighbour
    NBORS(from,nborsCount) = to;
    nborsCount++;
    numNbors[from] = nborsCount;

    // Success
    return 0;

}

void SimulationMesh::updateDist(int from, int to, double distance)
{
    if(distance <= 0.0)
    {
        qDebug() << "ERROR: A distance less or equal to zero is enqountered between node " \
                << from << " and " << to << ". Replacing with default value: 1.0.";
        distance = 1.0;
    }
    for (int i = 0; i < numNbors[from]; i++)
    {
        if (NBORS(from,i) == to)
        {
            DISTS(from,i) = distance;
            // get out of loop?
        }
    }
    for (int i = 0; i < numNbors[to]; i++)
    {
        if (NBORS(to,i) == from)
        {
            DISTS(to,i) = distance;
            // get out of loop?
        }
    }
    // should I do error-checking?
}


double SimulationMesh::getNborDist(int fromNode, int toNode)
{

    for (int i = 0; i < numNbors[fromNode]; i++)
    {
        if (NBORS(fromNode,i) == toNode)
        {
            return DISTS(fromNode,i);
        }
    }
}

double SimulationMesh::getNodeTemp(int node)
{
    return nodeTemps[node];
}

double SimulationMesh::getNodeFuel(int node)
{
    return nodeFuels[node];
}

bool SimulationMesh::getNodeIgn(int node)
{
    return nodeIgn[node]    ;

}



