#include "StdAfx.h"
#include "BurningObject.h"

BurningObject::BurningObject(GLfloat* vertexPositions, GLfloat* vertexNormals, GLfloat* vertexColors, 
							 GLfloat* vertexTexcoords, GLfloat* vertexDispl,
							 GLint* facelookuptable, GLint numVertices, GLint numDrawVerts, 
							 const char* vertShaderFilename, const char* fragShaderFilename, const char* textureFilename, 
							 Mesh* simMesh)
{
    loadIdMatrix4x4f(m_modelMatrix);

	this->numDrawVerts = numDrawVerts;
	initOpenCL();
	loadShaderProgram(vertShaderFilename, fragShaderFilename);
	makeVertexBuffers(vertexPositions, vertexNormals, vertexColors, vertexTexcoords, vertexDispl, facelookuptable, numVertices);
	initTexture(textureFilename);
	delete[] vertexPositions;
	delete[] vertexNormals;
	delete[] vertexColors;
	delete[] vertexTexcoords;
	delete[] facelookuptable;
	initCLBuffers(simMesh);


}

inline void clCheckErr(cl_int err, const char * name)
{
    if (err != CL_SUCCESS) {
        std::cerr << "ERROR: " << name
                 << " (" << err << ")" << std::endl;
		pausePressEnter();
        exit(EXIT_FAILURE);
	}
}

BurningObject::~BurningObject(void)
{
	cl_int status;

    status = clReleaseKernel(clTempKern);
    clCheckErr(status,"Error: In clReleaseKernel \n");

    //status = clReleaseKernel(clCopyKern);
    //clCheckErr(status,"Error: In clReleaseKernel \n");

    status = clReleaseKernel(clColorKern);
    clCheckErr(status,"Error: In clReleaseKernel \n");
	
    status = clReleaseKernel(clWindKern);
    clCheckErr(status,"Error: In clReleaseKernel \n");

    status = clReleaseProgram(clProgram);
    clCheckErr(status,"Error: In clReleaseProgram\n");

	for (int i = 0; i < NUM_IGNBUFFS; i++)
	{
		status = clReleaseMemObject(clIgnbuff[i]);
		clCheckErr(status,"Error: In clReleaseMemObject (inputBuffer)\n");
	}

	status = clReleaseMemObject(clTempsbuff);
    clCheckErr(status,"Error: In clReleaseMemObject (outputBuffer)\n");

	status = clReleaseMemObject(clFuelsbuff);
    clCheckErr(status,"Error: In clReleaseMemObject (outputBuffer)\n");

	status = clReleaseMemObject(clNborsbuff);
    clCheckErr(status,"Error: In clReleaseMemObject (outputBuffer)\n");

	status = clReleaseMemObject(clNborCountsbuff);
    clCheckErr(status,"Error: In clReleaseMemObject (outputBuffer)\n");

	status = clReleaseMemObject(clDistbuff);
    clCheckErr(status,"Error: In clReleaseMemObject (outputBuffer)\n");

	status = clReleaseMemObject(clMappingbuff);
    clCheckErr(status,"Error: In clReleaseMemObject (outputBuffer)\n");

    status = clReleaseCommandQueue(clQueue);
    clCheckErr(status,"Error: In clReleaseCommandQueue\n");

    status = clReleaseContext(clCtxt);
    clCheckErr(status,"Error: In clReleaseContext\n");

	delete clDevices;
}




/* IMPORTANT: 
 * This method creates the opencl context, and must be called AFTER the opengl context is created, 
 * but BEFORE any resources (such as buffer objects) that are to be shared between opengl and opencl are created!
 * If not, clEnqueueAcquireGLObjects() gives 'memory access violation'. 
 * From http://developer.amd.com/support/KnowledgeBase/Lists/KnowledgeBase/DispForm.aspx?ID=84:
 * "To use shared resources, the OpenGL application must first create an OpenGL context and then an OpenCL context. 
 * All resources created after the OpenCL context has been created can be shared between OpenGL and OpenCL. 
 * If resources are allocated before the OpenCL context is created, they cannot be shared between OpenGL and OpenCL."
 */
void BurningObject::initOpenCL(void)
{
	cl_int err;
	// Make cl context from gl context. OBS: Following uses wgl directly - WINDOWS ONLY!
	HGLRC hrc = wglGetCurrentContext();
	HDC hDC = wglGetCurrentDC();
	if (!hrc || !hDC)
		std::cout << "Something wrong with getting the context or dc in BurningObject::initTemperatureBuffers!" << std::endl;
	// Production maaten aa gjoere det paa (http://developer.amd.com/support/KnowledgeBase/Lists/KnowledgeBase/DispForm.aspx?ID=71):
	/*
     * Have a look at the available platforms and pick either
     * the AMD one if available or a reasonable default.
     */
    cl_uint numPlatforms;
    cl_platform_id platform = NULL;
    err = clGetPlatformIDs(0, NULL, &numPlatforms);
    clCheckErr(err, "clGetPlatformIDs failed(1).");
    if (0 < numPlatforms) 
    {
        cl_platform_id* platforms = new cl_platform_id[numPlatforms];
        err = clGetPlatformIDs(numPlatforms, platforms, NULL);
		clCheckErr(err, "clGetPlatformIDs failed(2).");
        for (unsigned i = 0; i < numPlatforms; ++i) 
        {
            char pbuf[100];
            err = clGetPlatformInfo(platforms[i],
                                       CL_PLATFORM_VENDOR,
                                       sizeof(pbuf),
                                       pbuf,
                                       NULL);
			clCheckErr(err, "clGetPlatformInfo failed.");
            platform = platforms[i];
            if (!strcmp(pbuf, "Advanced Micro Devices, Inc.")) 
            {
				std::cout << "AMD device found." << std::endl;
                break;
            }
        }
        delete[] platforms;
    }
    /*
     * If we could find our platform, use it. Otherwise pass a NULL and get whatever the
     * implementation thinks we should be using.
     */
	cl_context_properties props[7] = 
	{CL_GL_CONTEXT_KHR, (cl_context_properties) hrc, CL_WGL_HDC_KHR, (cl_context_properties) hDC,
        CL_CONTEXT_PLATFORM, 
        (cl_context_properties)platform,  NULL};
    /* Use NULL for backward compatibility */
    cl_context_properties* cprops = NULL;
	if (platform == NULL)
	{
		std::cout << "Warning, did not find a platform, using NULL" << std::endl;
	} else {
		cprops = props;
	}
    clCtxt = clCreateContextFromType(
                  cprops,
                  CL_DEVICE_TYPE_GPU,
                  NULL,
                  NULL,
                  &err);

	clCheckErr(err, "Context::Context()");


	size_t cb;
	clGetContextInfo(clCtxt, CL_CONTEXT_DEVICES, 0, NULL, &cb);
	clDevices = (cl_device_id*) malloc(cb);
	clGetContextInfo(clCtxt, CL_CONTEXT_DEVICES, cb, clDevices, NULL);

}

#define ARRAYLENGTH 10
void BurningObject::initCLBuffers(Mesh* simMesh)
{
	GLfloat deltaT = 0.0f;

	// can probably drop one of the following....
	clWorkSize[0] = simMesh->numNodes;
	this->numNodes = simMesh->numNodes;

	// create and initialize a node-ignited table
	// this should be a boolean, but I couldnt get it to work then :(
	cl_int* tmpNodeIgn = new cl_int[simMesh->numNodes];
	memset(tmpNodeIgn, 0, simMesh->numNodes*sizeof(cl_int));
	tmpNodeIgn[100] = 1; //ignite a random node
	tmpNodeIgn[10] = 1; //ignite a random node

	// create and initialize a link speed table
	cl_float* tmpLinkSpeed = new cl_float[simMesh->numNodes*simMesh->numMaxNbors];
	memset(tmpLinkSpeed, 0, simMesh->numNodes*simMesh->numMaxNbors*sizeof(cl_float));

	// initialize wind vector
	windVec[0] = 0.0;
	windVec[1] = 0.0;
	windVec[2] = 1.0;


	// read the kernel source from file
	GLchar* programsource = readFile("temperature_kernels.cl", 1000000);
    cl_int err;

	clProgram = clCreateProgramWithSource(clCtxt, 1, (const char**) &programsource, NULL, &err);//(strlen(programsource)+1)*sizeof(GLchar), &err);
	clCheckErr(err, "BurningObject::initCLBuffers: create program");
	err = clBuildProgram(clProgram, 1, clDevices, NULL, NULL, NULL);
	clCheckErr(err, "BurningObject::initCLBuffers: build program");

	std::cout << "Build program succeeded." << std::endl;
	//pausePressEnter();
	//exit(1);
	
	clTempKern = clCreateKernel(clProgram, "updatetemps", &err);
    clCheckErr(err, "BurningObject::initCLBuffers: create update kernel");
	//clCopyKern = clCreateKernel(clProgram, "copybacktemps", &err);
    //clCheckErr(err, "BurningObject::initCLBuffers: create copy kernel");
	clColorKern = clCreateKernel(clProgram, "updatecolors", &err);
    clCheckErr(err, "BurningObject::initCLBuffers: create color kernel");
	clWindKern = clCreateKernel(clProgram, "updatewind", &err);
    clCheckErr(err, "BurningObject::initCLBuffers: create wind kernel");
	
	clTempsbuff		= clCreateBuffer(clCtxt, CL_MEM_READ_WRITE | CL_MEM_COPY_HOST_PTR, 
		simMesh->numNodes*sizeof(GLfloat), simMesh->nodeTemperatures, &err);
    clCheckErr(err, "BurningObject::initCLBuffers: create temp buff 0");
	clIgnbuff[0]	= clCreateBuffer(clCtxt, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR, 
		simMesh->numNodes*sizeof(cl_int), tmpNodeIgn, &err);
    clCheckErr(err, "BurningObject::initCLBuffers: create temp buff 1");
	clIgnbuff[1]	= clCreateBuffer(clCtxt, CL_MEM_WRITE_ONLY | CL_MEM_COPY_HOST_PTR, 
		simMesh->numNodes*sizeof(cl_int), tmpNodeIgn, &err);
    clCheckErr(err, "BurningObject::initCLBuffers: create ignited buff");
	clFuelsbuff		= clCreateBuffer(clCtxt, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR, 
		simMesh->numNodes*sizeof(GLfloat), simMesh->nodeFuels, &err);
    clCheckErr(err, "BurningObject::initCLBuffers: create fuels buff");
	clFuelremsbuff	= clCreateBuffer(clCtxt, CL_MEM_READ_WRITE | CL_MEM_COPY_HOST_PTR, 
		simMesh->numNodes*sizeof(GLfloat), simMesh->nodeFuels, &err);
    clCheckErr(err, "BurningObject::initCLBuffers: create fuel-remaining buff");
	clNborsbuff		= clCreateBuffer(clCtxt, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR, 
		simMesh->numNodes*simMesh->numMaxNbors*sizeof(GLint), simMesh->neighbours, &err);
    clCheckErr(err, "BurningObject::initCLBuffers: create nbors buff");
	clNborCountsbuff= clCreateBuffer(clCtxt, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR, 
		simMesh->numNodes*sizeof(GLint), simMesh->nborCounts, &err);
    clCheckErr(err, "BurningObject::initCLBuffers: create nborcounts buff");
	clDistbuff		= clCreateBuffer(clCtxt, CL_MEM_READ_WRITE | CL_MEM_COPY_HOST_PTR, 
		simMesh->numNodes*simMesh->numMaxNbors*sizeof(GLfloat), simMesh->distances, &err);
    clCheckErr(err, "BurningObject::initCLBuffers: create dist buff");
	clLinkDirbuff	= clCreateBuffer(clCtxt, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR, 
		simMesh->numNodes*simMesh->numMaxNbors*sizeof(GLfloat)*3, simMesh->linkDirs, &err);
    clCheckErr(err, "BurningObject::initCLBuffers: create dist buff");
	clWindVecbuff	= clCreateBuffer(clCtxt, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR, 
		3*sizeof(GLfloat), windVec, &err);
    clCheckErr(err, "BurningObject::initCLBuffers: create dist buff");
	clLinkSpeedbuff	= clCreateBuffer(clCtxt, CL_MEM_READ_WRITE | CL_MEM_COPY_HOST_PTR, 
		simMesh->numNodes*simMesh->numMaxNbors*sizeof(GLfloat), tmpLinkSpeed, &err);
    clCheckErr(err, "BurningObject::initCLBuffers: create dist buff");
	clMappingbuff	= clCreateBuffer(clCtxt, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR, 
		simMesh->numNodes*sizeof(GLint), simMesh->vertexMap, &err);
    clCheckErr(err, "BurningObject::initCLBuffers: create mapping buff");
	clSharedbuff[SHAREDBUFF_COLOR] = clCreateFromGLBuffer(clCtxt, CL_MEM_READ_WRITE, vertexBuffersObject[GLOW_BUFFER], &err);
    clCheckErr(err, "BurningObject::initCLBuffers: create shared color buff");
	clSharedbuff[SHAREDBUFF_DISPL] = clCreateFromGLBuffer(clCtxt, CL_MEM_READ_WRITE, vertexBuffersObject[DISPL_BUFFER], &err);
    clCheckErr(err, "BurningObject::initCLBuffers: create shared displ buff");

 
	err = clSetKernelArg(clTempKern, ARGNO_TEMPKERN_TEMPS, sizeof(clTempsbuff), &clTempsbuff);
    clCheckErr(err, "clUpdateKern::setArg(0)");
	err = clSetKernelArg(clTempKern, ARGNO_TEMPKERN_READIGN, sizeof(clIgnbuff[0]), clIgnbuff);//&clIgnbuff[0]);
    clCheckErr(err, "clUpdateKern::setArg(1)");
	err = clSetKernelArg(clTempKern, ARGNO_TEMPKERN_WRITEIGN, sizeof(clIgnbuff[1]), clIgnbuff+1);//&clIgnbuff[1]);
    clCheckErr(err, "clUpdateKern::setArg(2)");
	err = clSetKernelArg(clTempKern, ARGNO_TEMPKERN_FUEL, sizeof(clFuelremsbuff), &clFuelremsbuff);
    clCheckErr(err, "clUpdateKern::setArg(2)");
	err = clSetKernelArg(clTempKern, ARGNO_TEMPKERN_NBORSWIDTH, sizeof(GLint), &(simMesh->numMaxNbors));
    clCheckErr(err, "clUpdateKern::setArg(3)");
	err = clSetKernelArg(clTempKern, ARGNO_TEMPKERN_NBORS, sizeof(clNborsbuff), &clNborsbuff);
    clCheckErr(err, "clUpdateKern::setArg(4)");
	err = clSetKernelArg(clTempKern, ARGNO_TEMPKERN_DISTANCES, sizeof(clDistbuff), &clDistbuff);
    clCheckErr(err, "clUpdateKern::setArg(5)");
	err = clSetKernelArg(clTempKern, ARGNO_TEMPKERN_NBORCOUNTS, sizeof(clNborCountsbuff), &clNborCountsbuff);
    clCheckErr(err, "clUpdateKern::setArg(6)");
	err = clSetKernelArg(clTempKern, ARGNO_TEMPKERN_LINKSPEEDS, sizeof(clLinkSpeedbuff), &clLinkSpeedbuff);
    clCheckErr(err, "BurningObject::initCLBuffers: set windkernel argument linkspeeds");
	err = clSetKernelArg(clTempKern, ARGNO_TEMPKERN_DELTAT, sizeof(GLfloat), &deltaT);
    clCheckErr(err, "clUpdateKern::setArg(7)");
	
	/*
	err = clSetKernelArg(clCopyKern, ARGNO_COPYKERN_READIGN, sizeof(clIgnbuff[0]), clIgnbuff);
    clCheckErr(err, "clCopyKern::setArg(0)");
	err = clSetKernelArg(clCopyKern, ARGNO_COPYKERN_WRITEIGN, sizeof(clIgnbuff[1]), clIgnbuff+1);
    clCheckErr(err, "clCopyKern::setArg(1)");
	*/

	
	err = clSetKernelArg(clColorKern, ARGNO_COLKERN_READTEMPS, sizeof(clTempsbuff), &clTempsbuff);
    clCheckErr(err, "clColorKern::setArg(0)");
	err = clSetKernelArg(clColorKern, ARGNO_COLKERN_FUELINIT, sizeof(clFuelsbuff), &clFuelsbuff);
    clCheckErr(err, "clColorKern::setArg(1)");
	err = clSetKernelArg(clColorKern, ARGNO_COLKERN_FUELREM, sizeof(clFuelremsbuff), &clFuelremsbuff);
    clCheckErr(err, "clColorKern::setArg(1)");
	err = clSetKernelArg(clColorKern, ARGNO_COLKERN_MAPPING, sizeof(clMappingbuff), &clMappingbuff);
    clCheckErr(err, "clColorKern::setArg(2)");
	err = clSetKernelArg(clColorKern, ARGNO_COLKERN_COLORS, sizeof(clSharedbuff[SHAREDBUFF_COLOR]), clSharedbuff+SHAREDBUFF_COLOR); // pointer arithemetic
    clCheckErr(err, "clColorKern::setArg(3)");
	err = clSetKernelArg(clColorKern, ARGNO_COLKERN_DISPL, sizeof(clSharedbuff[SHAREDBUFF_DISPL]), clSharedbuff+SHAREDBUFF_DISPL);// pointer arithemetic
    clCheckErr(err, "clColorKern::setArg(4)");

	err = clSetKernelArg(clWindKern, ARGNO_WINDKERN_LINKDIRS, sizeof(clLinkDirbuff), &clLinkDirbuff);
	clCheckErr(err, "BurningObject::initCLBuffers: set windkernel argument linkdirs");
	err = clSetKernelArg(clWindKern, ARGNO_WINDKERN_NBORSWIDTH, sizeof(GLint), &(simMesh->numMaxNbors));
    clCheckErr(err, "BurningObject::initCLBuffers: set windkernel argument nborswidth");
	err = clSetKernelArg(clWindKern, ARGNO_WINDKERN_WINDVEC, sizeof(clWindVecbuff), &clWindVecbuff);
    clCheckErr(err, "BurningObject::initCLBuffers: set windkernel argument windvec)");
	err = clSetKernelArg(clWindKern, ARGNO_WINDKERN_LINKSPEEDS, sizeof(clLinkSpeedbuff), &clLinkSpeedbuff);
    clCheckErr(err, "BurningObject::initCLBuffers: set windkernel argument linkspeeds");

	clQueue = clCreateCommandQueue(clCtxt, clDevices[0], 0, &err);
    clCheckErr(err, "CommandQueue::CommandQueue()");
	
	
	delete[] programsource;
	delete[] tmpNodeIgn;

	//for (int times = 0; times < 10; times++)
	//{
	//	updateTemperatures();
	//}

}


void BurningObject::updateTemperatures(GLfloat deltaT)
{
	//Debug: GLfloat * outH = new GLfloat[elementCounts.numPositions];
	//Debug: GLfloat * outColors = new GLfloat[elementCounts.numVertices * 3];

	cl_int err;
	cl_event rdyevent[1];
	cl_event rdyevent2[1];	// it doesnt seem like I can use the same event as wait event and release event, 
							// so I make two and use them alternately.
	
	// update deltaT
	err = clSetKernelArg(clTempKern, ARGNO_TEMPKERN_DELTAT, sizeof(GLfloat), &deltaT);
    clCheckErr(err, "Kernel::setArg(7) - update deltaT");
	
	//colorbuff and displbuff needs to be acquired:
	err = clEnqueueAcquireGLObjects(clQueue, NUM_SHAREDBUFFS, clSharedbuff, 0, NULL, rdyevent);
    clCheckErr(err, "clEnqueueAcquireGLObjects()");

	//clWaitForEvents(1,rdyevent);

	err = clEnqueueNDRangeKernel(clQueue, clTempKern, 1, NULL, clWorkSize, NULL, 1,rdyevent, rdyevent2);
	clCheckErr(err, "BurningObject::updateTemperatures enque update kernel");
	//clWaitForEvents(1,rdyevent);

	/*err = clEnqueueNDRangeKernel(clQueue, clCopyKern, 1, NULL, clWorkSize, NULL, 0, NULL, rdyevent);
	clCheckErr(err, "ComamndQueue::enqueueNDRangeKernel()");
	clWaitForEvents(1,rdyevent);
	*/
	err = clEnqueueCopyBuffer(clQueue, clIgnbuff[1], clIgnbuff[0], 0, 0, numNodes*sizeof(cl_int), 1,rdyevent2, rdyevent);
	clCheckErr(err, "BurningObject::updateTemperatures: copy ign buffer");
	//clWaitForEvents(1,rdyevent);
	

	//Debug: err = clEnqueueReadBuffer(clQueue, clTempsbuff1, CL_TRUE, 0, elementCounts.numPositions * sizeof(GLfloat), outH, 0, NULL, NULL);
	// clCheckErr(err, "ComamndQueue::enqueueReadBuffer()");
	//for (int i = 0; i < 400; i++)//elementCounts.numPositions; i++)
	//	std::cout << outH[i] << ", ";
	//std::cout << std::endl;
	
	//std::cout << "Copying to colorbuffer..." << std::endl;
	err = clEnqueueNDRangeKernel(clQueue, clColorKern, 1, NULL, clWorkSize, NULL, 1,rdyevent, rdyevent2);
	clCheckErr(err, "BurningObject::updateTemperatures: enque color kernel");
	//clWaitForEvents(1,rdyevent);  
	//std::cout << "Finished copying colors!" << std::endl;

	//Debug: err = clEnqueueReadBuffer(clQueue, clColorbuff, CL_TRUE, 0, elementCounts.numVertices * 3 * sizeof(GLfloat), outColors, 0, NULL, NULL);
	//clCheckErr(err, "ComamndQueue::enqueueReadBuffer()");
	//for (int i = 0; i < 400; i++)//elementCounts.numPositions; i++)
	//	std::cout << outColors[i] << ", ";
	//std::cout << std::endl;

	//colorbuff and displbuff needs to be released:
	err = clEnqueueReleaseGLObjects(clQueue, NUM_SHAREDBUFFS, clSharedbuff, 1,rdyevent2, rdyevent);
	clCheckErr(err, "BurningObject::updateTemperatures: release gl buffers");

	clWaitForEvents(1,rdyevent);

	//Debug: delete [] outH;
	//Debug: delete [] outColors;

}

void BurningObject::updateWind(GLfloat wind_x, GLfloat wind_y, GLfloat wind_z)
{
	cl_int err;
	cl_event rdyevent[1];
	cl_event rdyevent2[1];

	windVec[0] = wind_x;
	windVec[1] = wind_y;
	windVec[2] = wind_z;

	err = clEnqueueWriteBuffer(clQueue, clWindVecbuff, CL_TRUE, 0, 3*sizeof(cl_float), windVec, 0, NULL, rdyevent);
	clCheckErr(err, "BurningObject::updateTemperatures: copy ign buffer");

	err = clEnqueueNDRangeKernel(clQueue, clWindKern, 1, NULL, clWorkSize, NULL, 1, rdyevent, rdyevent2);
	clCheckErr(err, "BurningObject::updateTemperatures enque update kernel");

	clWaitForEvents(1,rdyevent2);


}



