#include "StdAfx.h"
#include "ObjData.h"
#pragma warning(once: 4996)

#define MAX_FILE_LENGTH 1000000
#define MAX_LINE_LENGTH 1000
ObjData::ObjData(const char* objFileName, GLfloat scale, GLboolean makeMesh)
{
	// Streams and buffers
	std::ifstream instream;
	char* line = new char[MAX_LINE_LENGTH];
	instream.open(objFileName);
	if(!instream) {
		std::cout << "Cannot open file: " << objFileName << "." << std::endl;
		pausePressEnter();
		return;
	}

	instream.clear();					// forget if we've hit the end of file
	instream.seekg(0, std::ios::beg);   // move to the start of the file

	std::vector<RawFaceDataPointers> rawFaceIndices = loadDataFromObj(&instream, MAX_LINE_LENGTH, scale);
	instream.close();
	
	std::cout << "Filled object with: " << std::endl << "  " << \
		vertPositions.size() << " pos. coordinates, " << \
		vertNormals.size() << " norms, " << \
		vertTexture.size() << " texcoords " << \
		numFaces << " faces." << std::endl;

	checkPositions(1000, 0.001F);
	makeLookupTables(&rawFaceIndices);
	checkNumVertices();
	checkForUnusedPositions();
	makeNodeOffsetTable();
	if(makeMesh){
		makeNeighbourMatrix();//faceIndices);
		makeDistMatrix();
	}


}

ObjData::~ObjData(void)
{
	delete[] nodeOffsets;
	delete[] nborCounts;
}


inline GLboolean vertexEqual(VertexDataPointers a, VertexDataPointers b)
{
	return a.texIdx == b.texIdx && \
		a.normalIdx == b.normalIdx && \
		a.colorIdx == b.colorIdx;
}
inline GLboolean rawVertexEqual(RawFaceDataPointers a, VertexDataPointers b)
{
	return a.texIdx == b.texIdx && \
		a.normalIdx == b.normalIdx && \
		a.colorIdx == b.colorIdx;
}


// Making a vector grouped on vertex positions. 
// Takes the face-indices data from the .obj-file and converts it to a two-step lookup:
// - One set of indices into the vertPosition, vertNormal and vertColor tables for each unique vertex, 
// and one set of indices into this vertex lookup table for constructing faces (a face lookup table).
// This is needed because OpenGL cannot take separate indices for position, normal and color for each vertex.
void ObjData::makeLookupTables(std::vector<RawFaceDataPointers>* rawFacesTable) // using pointer because it is so big...
{

	// keep track of number of vertices inserted, 
	// in case it differs from the number found in the .obj-file
	// (because of duplicate vertices or unused vert positions)
	numVertices = 0;

	int vertHits = 0;

	VertexIndex emptyvi = {-1,-1};

	faces.assign(numFaces*numVertsPrFace, emptyvi);

	// fill vertices with negative vertex data pointers indicating they're empty.
	VertexDataPointers emptyvdp = {-1,-1,-1};
	std::vector<VertexDataPointers> v(1,emptyvdp);
	vertices.assign(vertPositions.size(), v);



	// for each vertex in the raw faces table
	for (GLint i = 0; i < (numFaces*numVertsPrFace); i++)
	{
		int pos = rawFacesTable->at(i).positionIdx;
		// check if the first position is free
		if (vertexEqual(vertices[pos][0], emptyvdp))
		{
			// replace the empty vertex
			vertices[pos][0].normalIdx = rawFacesTable->at(i).normalIdx;
			vertices[pos][0].texIdx = rawFacesTable->at(i).texIdx;
			vertices[pos][0].colorIdx = rawFacesTable->at(i).colorIdx;
			// add the index to the faceLookup, and continue
			faces[i].pos = pos;
			faces[i].no = 0;
			numVertices++;
			continue;
		}
		// check if the lookup-table already contains this vertex, 
		// by searching through the vertices with this position-index
		// (As there doesnt seem to be a vector-find function, I must do this manually)
		GLboolean found = GL_FALSE;
		for (unsigned int num = 0; num < vertices[pos].size(); num++)
		{
			if(rawVertexEqual(rawFacesTable->at(i), vertices[pos][num]))
			{
				// found an equal vertex-data-pointer
				found = GL_TRUE;
				vertHits++;
				faces[i].pos = pos;
				faces[i].no = num;
				break;
			}
		}
		// skip to the next face vertex if we found an equal vertex-data-pointer above.
		if (found) continue;

		// if none of the above, append to the end of the lookuptable
		VertexDataPointers v = {rawFacesTable->at(i).normalIdx, rawFacesTable->at(i).texIdx, rawFacesTable->at(i).colorIdx};
		vertices[pos].push_back(v);
		faces[i].pos = pos;
		faces[i].no = vertices[pos].size()-1;
		numVertices++;
	}

	std::cout << "Created " << numVertices << " vertices; " << vertHits << " vertex reuses (sum: " << 
		numVertices+vertHits << ", should be " << numFaces*numVertsPrFace << ")." << std::endl;
	std::cout << "Printing a sample of the vertex data..." << std::endl;

	// DEBUG: print the first X vertices.
	int printcount = 10;
	int pos = 0;
	int no = 0;
	int i = 0;
	while (i < printcount)
	{
		int maxNo = vertices[pos].size();
		std::cout << "  There are " << maxNo << " vertices on position " << pos << std::endl;
		//if (maxNo > 5)
		//	maxNo = 5;
		if (no >= maxNo){
			no = 0;
			pos++;
		}
		std::cout << "  Vertex no. (" << pos << ", " << no << ") has the following data: " << std::endl;
		printVertexData(pos,no);
		no++;
		i++;
		
	}

	// DEBUG: print X faceindices:
	printcount = 50;
	int printstart = 0;
	std::cout << "Printing sample of " << printcount << " face indices:" << std::endl << "  "; 
	for (i = printstart; i < printstart+printcount; i++)
	{
		std::cout << "(" << faces[i].pos << ", " << faces[i].no << "), ";
	}
	std::cout << "." << std::endl;




}

void ObjData::checkNumVertices(void){
	int count = 0;
	for(std::vector<std::vector<VertexDataPointers>>::iterator vertpos = vertices.begin();
			vertpos != vertices.end(); ++vertpos)
	{
		count += vertpos->size();
	}
	std::cout << "NumVertices-check: 'numVertices'=" << numVertices << " actual vertex count=" << count << std::endl;
	if (numVertices != count)
		std::cout << "  WARNING: numVertices differs from actual vertex count! Possibly because of unused vertex positions. Suggest you run uniquify.py." << std::endl;
}

void ObjData::checkForUnusedPositions(void){
	std::cout << "Unused pos-check..." << std::endl;
	VertexDataPointers emptyvertex = {-1,-1,-1};
	std::vector<int> unused;
	for(unsigned int vertpos = 0; vertpos < vertices.size(); vertpos++)
	{
		if(vertexEqual(vertices[vertpos][0], emptyvertex))
			if(vertices[vertpos].size() > 1)
				std::cout << "WARNING: Filled vertices trailing an empty vertex position in pos " << vertpos << std::endl;
			else
				unused.push_back(vertpos);
	}
	if (unused.size() > 0)
	{
		std::cout << "  found " << unused.size() << " unused positions:" << std::endl;
		for (unsigned int i = 0; i < unused.size(); i++)
			std::cout << unused[i] << ", ";
		std::cout << "." << std::endl;
	}else{
		std::cout << "  did not find any unused positions." << std::endl;
	}
	
}

void ObjData::printVertexData(int pos, int no)
{
	std::cout << "   Position: ";
	printVec3f(&(vertPositions[pos]));
	std::cout << std::endl << "   Normal: ";
	printVec3f(&(vertNormals[vertices[pos][no].normalIdx]));
	std::cout << std::endl << "   Color: ";
	printVec3f(&(vertColors[vertices[pos][no].colorIdx]));
	std::cout << "." << std::endl;

}

void ObjData::printVertexIndices(int pos, int no)
{
	std::cout << "   Pos idx: " << pos << 
		", Norm idx: " << vertices[pos][no].normalIdx <<
		", Color idx: " << vertices[pos][no].colorIdx << "." << std::endl;

}



#define VERTS_PR_FACE 3
// Fills out the neighbour matrix 'nbors'.
void ObjData::makeNeighbourMatrix()
{
	int err1count, err2count;
	err1count = err2count = 0;
	const int errthresh = 10;
	GLuint nboridx = 0;
	GLuint thisidx = 0;
	GLuint max_nbors_actual = 0;
	
	//initialize neighbours to an "illegal" neighbour (negative) to indicate they are not yet filled (empty/available)
	/*for (unsigned int i = 0; i < vertPositions.size(); i++)
	{
		std::vector<int> v(1,-1); // temporary vector with length 1 initialized to -1
		neighbours.push_back(v);
	}*/
	std::vector<int> v(1,-1);
	neighbours.assign(vertPositions.size(), v);

	// loop through the face-table and register all positions sharing a face as neighbours in the mesh.
	for (int f = 0; f < numFaces; f++)
	{
		int first = f * numVertsPrFace;

		// for each vertex of the current face
		for (int v = 0; v < numVertsPrFace; v++)
		{
			int thisidx = first+v; // buffered for debug purposes
			int thispos = faces[thisidx].pos;

			// for each of the other vertices in the face
			for (int n = 1; n < numVertsPrFace; n++)
			{
				int nboridx = first+((v+n)%numVertsPrFace); // buffered for debug purposes
				int nborpos = faces[nboridx].pos;
				GLboolean addnbor = GL_TRUE;
				// check if this-vertex' position and nbor-vertex' position is the same.
				if (thispos == nborpos)
				{
					addnbor = GL_FALSE;
					if(err1count <= errthresh){
						std::cout << "Warning: Vertex-pos " << thispos << \
							" met itself as a neighbour in face number " << f <<
							". It will not be added to it's own nborlist. " << 
							"This may (must?) indicate an infinitesimal triangle, and should be removed by running uniquify.py on the model data." << std::endl;
						std::cout << "  Vertex indices this: ";
						printVertexIndices(faces[thisidx].pos, faces[thisidx].no);
						std::cout << "  Vertex indices nbor: ";
						printVertexIndices(faces[nboridx].pos, faces[nboridx].no);
						err1count++;
					}
					if(err1count == errthresh)
					{
						std::cout << "This is the " << errthresh << "th warning - subsequent errors of this kind will be silently ignored." << std::endl; 
					}
				}
				//check if neighbours already
				for (unsigned int i = 0; i < neighbours[thispos].size(); i++)
				{
					if (neighbours[thispos][i]==nborpos)
					{
						addnbor = GL_FALSE;
						if(err2count <= errthresh){
							std::cout << "Warning: Vertex at pos " << thispos << " and " << nborpos << \
								" are already neibhbours. They will not be registered anew." << std::endl;
							err2count++;
						}
						break; // don't need to check the rest of the neighbours
					}
				}
				// if this is a new neighbour; add nborpos to thispos' neighbourlist
				if(addnbor){
					if (neighbours[thispos][0] < 0) // first check if thispos has no neighbours yet
						neighbours[thispos][0] = nborpos; //then we need to replace the current placeholder
					else
						neighbours[thispos].push_back(nborpos); //else we append it to the end of the list
				}
			}// end for each of the other vertices in the face
		}// end for each vertex of the current face
	}// end for each face

	// find neighbour counts for all nodes and max number of neighbours for any node
	// (this could be done ("innlemmet i") during the construction of "neighbours" for better performance.)
	numMaxNbors = 0;
	nborCounts = new GLint[vertPositions.size()];
	for (unsigned int node = 0; node < vertPositions.size(); node++)
	{
		int count = neighbours[node].size();
		nborCounts[node] = count;
		if(count > numMaxNbors)
			numMaxNbors = count;
	}

	// DEBUG: Print out a some random neighbours for testing.
	const GLuint testpos = vertPositions.size()/6;
	std::cout << "Random nbors-test:" << std::endl \
		<< "Node no. " << testpos \
		<< " has " << neighbours[testpos].size() << " neighbours:" << std::endl;
	for (unsigned int i = 0; i < neighbours[testpos].size(); i++)
	{
		std::cout << neighbours[testpos][i] << ", ";
	}
	std::cout << std::endl;
	//std::cout << "Nborswidth is: " << nborswidth << std::endl;

}


inline bool ObjData::dupCheck(int posIndex1, int posIndex2, GLfloat thresh)
{
return (
		fabsf(vertPositions[posIndex1].x - vertPositions[posIndex2].x) < thresh &&  
		fabsf(vertPositions[posIndex1].y - vertPositions[posIndex2].y) < thresh &&  
		fabsf(vertPositions[posIndex1].z - vertPositions[posIndex2].z) < thresh
		);
}

/**
  * arguments:
  * maxPos: maximum number of positions that will be checked. NB: Should never be above 50.000!
  * errThresh: max allowed error in distance when checking for equality.
  */
void ObjData::checkPositions(unsigned int maxPos, GLfloat errThresh){
	int errcount = 0;
	const int errthresh = 10;
	const float slingringsmonn = 0.0000001f;
	std::vector<std::pair<int,int>> duplist;
	std::cout << "Checking for duplicate position entries..." << std::endl;
	int numP = vertPositions.size();
	if (vertPositions.size() > maxPos)
	{
		numP = maxPos;
		std::cout << "  Warning: The object has " << vertPositions.size() << 
			" positions, but only the first " << maxPos << " will be checked." << std::endl;
	}
	for (int p = 0; p < (numP-1); p++)
	{
		for (int i = (p+1); i < numP; i++)
		{
			bool dup = dupCheck(p,i,slingringsmonn);
			if (errcount < errthresh && dup)
			{
				duplist.push_back(std::pair<int,int>(p,i));
				errcount++;
			}
		}
	}
	if (errcount > 0)
	{
		if (errcount < errthresh)
			std::cout << "  Found " << errcount << " possible duplicates:" << std::endl;
		else
			std::cout << "  Found many possible duplicates, here are the first " << errcount << ":" << std::endl;
		for (std::vector<std::pair<int,int>>::iterator it = duplist.begin();
			it != duplist.end();
			it++)
		{
			std::cout << "  " << it->first << " and " << it->second << std::endl;
		}
		std::cout << "  Suggest you use uniquify.py on your data." << std::endl;
	}


}


inline float ObjData::vertDistance(int posIndex1, int posIndex2)
{
	return sqrtf(
				powf((vertPositions[posIndex1].x - vertPositions[posIndex2].x), 2) + 
				powf((vertPositions[posIndex1].y - vertPositions[posIndex2].y), 2) + 
				powf((vertPositions[posIndex1].z - vertPositions[posIndex2].z), 2)
				);
}



void ObjData::makeDistMatrix(void)
{
	std::cout << "Making the distances matrix... " << std::endl;
		// initialize distance-matrix
	for (unsigned int i = 0; i < vertPositions.size(); i++)
	{
		std::vector<GLfloat> v(nborCounts[i],-1.0);
		distances.push_back(v);
	}

	int errcount = 0;
	const int errthresh = 20;

	//for each node
	for (unsigned int node = 0; node < vertPositions.size(); node++)
	{
		//for each of the node's neighbours
		for (unsigned int nborno = 0; nborno < neighbours[node].size(); nborno++)
		{
			int nborposidx = neighbours[node][nborno];
			distances[node][nborno] = vertDistance(node,nborposidx);
			if ((errcount <= errthresh) && (distances[node][nborno] <= 0.00001))
			{
				errcount++;
				std::cout << "Warning: Very short distance between neighbours " << \
					node << " and " << nborposidx << ": " << distances[node][nborno] << std::endl << "Their positions: (" << 
					vertPositions[node].x << "," << vertPositions[node].y << "," << vertPositions[node].z << ") and (" <<
					vertPositions[nborno].x << "," << vertPositions[nborno].y << "," << vertPositions[nborno].z << ")." << std::endl;
				if (errcount >= errthresh)
					std::cout << "Warning! More than " << errthresh << "errors! Skipping any remaining errors." << std::endl;
			}

		}

	}

	std::cout << "... distance matrix constructed!" << std::endl;

	const GLint testpos = vertPositions.size()/6;
	const GLint testnumbers = distances[testpos].size();
	std::cout << "Random test:" << std::endl \
		<< "Vertex no. " << testpos \
		<< " has " << testnumbers << " neighbours with distances: " << std::endl;
	for (int i = 0; i < testnumbers; i++)
	{
		std::cout << distances[testpos][i] << ", ";
	}
	std::cout << std::endl;

}

inline float vecLength(Vec3f vec)
{
	return sqrtf(
				powf(vec.x, 2) + 
				powf(vec.y, 2) + 
				powf(vec.z, 2)
				);
}

void ObjData::checkNormals(void)
{
	std::cout << "Checking that normals are equal for each vertex on each position." << std::endl;
	std::vector<int> unequalpos;
	for (unsigned int pos = 0; pos < vertices.size(); pos++)
	{
		unsigned int numV = vertices[pos].size();
		if (numV > 1) // only point in checking if there is more than one vertex at this position
		{
			for (unsigned int no = 0; no < numV; no++)
			{
				for (unsigned int no2 = no; no2 < numV; no2++)
				{
					Vec3f n1 = vertNormals[vertices[pos][no].normalIdx];
					Vec3f n2 = vertNormals[vertices[pos][no2].normalIdx];
					if(!vec3Equal(n1,n2, 0.00001f))
					{
						unequalpos.push_back(pos);
					}
				}
			}

		}
	}
	int printcount = 3;
	if (unequalpos.size() < printcount)
		printcount = unequalpos.size();
	std::cout << "  Found " << unequalpos.size() << " unequal normals." << std::endl <<
		"  Printing the first " << printcount << ":" <<std::endl;
	
	for (int i = 0; i < printcount; i++)
	{
		int pos = unequalpos[i];
		std::cout << "  ## Vertices at position " << pos << ": ##" << std::endl;
		for (int no = 0; no < vertices[pos].size(); no++)
		{
			printVertexData(pos,no);
		}
	}

	std::cout << "Checking normal lengths are 1." << std::endl;
	std::vector<std::pair<int,float>> wronglen;
	const float errthresh = 0.0001;
	for (unsigned int n = 0; n < vertNormals.size(); n++)
	{
		float l = vecLength(vertNormals[n]);
		if((l - 1.0 - errthresh) > 0.0f)
			wronglen.push_back(std::pair<int,float>(n,l));
	}
	int printcount2 = 5;
	if (wronglen.size() < printcount2)
		printcount2 = wronglen.size();
	std::cout << "  Found " << wronglen.size() << " normals with lengths different from 1.0." <<
		"  Printing the first " << printcount2 << ":" <<std::endl;
	
	for (int i = 0; i < printcount2; i++)
	{
		Vec3f v = vertNormals[wronglen[i].first];
		std::cout << "  ## Normal " << wronglen[i].first << "( len: " << wronglen[i].second << "): [" <<
			v.x << ", " << v.y << ", " << v.z << "]" << std::endl;
	}
	
	

}



#define ONE_INDEXED_OBJ 1
// Loads the vertex data from the obj file (stores them in the vert-tables)
// Returns the "raw" faces table that is to be converted into the two-step lookup tables.
// Courtesy of Nate Robins, 1997, 2000 - glm.cpp (heavily modified)
std::vector<RawFaceDataPointers> ObjData::loadDataFromObj(std::ifstream* instream, int maxLineLength, GLfloat scale //, std::set<RawPos>* uniquePosStrings, std::vector<int>* posIdxTranslation
		)
{
	std::vector<RawFaceDataPointers> rawFaceIndices;
	char* line = new char[maxLineLength];
	numVertsPrFace = 3; //hardcoding this for now - only consider triangulated objects
	
	GLint pidx, nidx, tidx;
	pidx = nidx = tidx = 0;

	while(!instream->eof()) {
		*instream >> std::setw(maxLineLength) >> line;
		switch(line[0]) {
			case '#':               // comment
				instream->getline(line, maxLineLength);
				break;
			case 'v':               // v, vn, vt
				switch(line[1]) {
					case '\0':          // position
						Vec3f p;
						*instream >> p.x;
						*instream >> p.y;
						*instream >> p.z;
						p.x *= scale;
						p.y *= scale;
						p.z *= scale;
						vertPositions.push_back(p);
						break;
					case 'n':           // normal
						Vec3f n;
						*instream >> n.x;
						*instream >> n.y;
						*instream >> n.z;
						vertNormals.push_back(n);
						break;
					case 't':           // texcoord				
						Vec3f t;
						*instream >> t.x;
						*instream >> t.y;
						*instream >> t.z;
						t.x *= scale;
						t.y *= scale;
						t.z *= scale;
						vertTexture.push_back(t);
						break;
					case 'a':           // tangents					
						Vec3f a;
						*instream >> a.x;
						*instream >> a.y;
						*instream >> a.z;
						//silently discard
						break;
					case 'r':           // radiuses (radii?)						
						Vec3f r;
						*instream >> r.x;
						*instream >> r.y;
						*instream >> r.z;
						r.x *= scale;
						r.y *= scale;
						r.z *= scale;
						vertRadius.push_back(r.x);
						break;
				}//end switch
				break;
			case 'u': // ?
				instream->getline(line, maxLineLength);
				break;
			case 'g':               // group
				instream->getline(line, maxLineLength);
				break;
			case 'f':               // face
				for (int i = 0; i < numVertsPrFace; i++) //repeat for the three vertices of the triangle
				{
					*instream >> std::setw(maxLineLength) >> line;
					pidx = nidx = tidx = ONE_INDEXED_OBJ; 
					// can be one of %d, %d//%d, %d/%d, %d/%d/%d %d//%d 
					if (strstr(line, "//")) {
						// v//n
						sscanf(line, " %d//%d ", &pidx, &nidx);
					} else if (sscanf(line, " %d/%d/%d ", &pidx, &tidx, &nidx) > 0) {
						// v/t/n
					} else if (sscanf(line, " %d/%d ", &pidx, &tidx) > 0) {
						// v/t
					} else if (sscanf(line, " %d ", &pidx) > 0) {
						// v/t
					} else {
						std::cerr << "ERROR: Found face with unknown vertex index format in .obj-file!" << std::endl;
					}//end if else
					RawFaceDataPointers f;
					f.positionIdx = pidx - ONE_INDEXED_OBJ - JORGENTREE_BUGFIX; //trekker fra 1 for aa rette opp i feil i jorgens tremodell
					f.texIdx = tidx - ONE_INDEXED_OBJ;
					f.normalIdx = nidx - ONE_INDEXED_OBJ;
					f.colorIdx = 1 - ONE_INDEXED_OBJ;
					rawFaceIndices.push_back(f);

				}//end for
				//faceIdx++;
				break;
                
			default:
				instream->getline(line, maxLineLength);
				break;
		} //end switch

	}//end while

	delete [] line;

	numFaces = rawFaceIndices.size()/numVertsPrFace;

	Vec3f onevec = {0.1f, 0.9f, 0.1f};
	if (vertPositions.empty()) 
	{
		vertPositions.push_back(onevec);
		//vertPositions[0] = {1.0f, 1.0f, 1.0f}
		std::cout << "Warning: No positions found - creating a dummy 1-length list." << std::endl;
	}
	
	if (vertNormals.empty()) 
	{
		vertNormals.push_back(onevec);
		//vertNormals[0] = {1.0f, 1.0f, 1.0f}
		std::cout << "Warning: No normals found - creating a dummy 1-length list." << std::endl;
	}
	
	if (vertColors.empty()) 
	{
		vertColors.push_back(onevec);
		//vertTexCoords[0] = {1.0f, 1.0f, 1.0f}
		std::cout << "Warning: No colors found - creating a dummy 1-length list." << std::endl;
	}
	
	if (vertTexture.empty()) 
	{
		vertTexture.push_back(onevec);
		//vertTexCoords[0] = {1.0f, 1.0f, 1.0f}
		std::cout << "Warning: No texture coordinates found - creating a dummy 1-length list." << std::endl;
	}

	return rawFaceIndices;

}


// finds the ending-offsets for each row/position/node group in "vertices"
// used a.o. to translate from the (pos,no) format of the internal face lookup table, 
// to the plain format of the external one. 
void ObjData::makeNodeOffsetTable(void){
	nodeOffsets = new GLint[vertPositions.size()];
	GLint accumOffs = 0; // the accumulated offset
	for (unsigned int pos = 0; pos < vertPositions.size(); pos++)// for each position, except the last	
	{
		accumOffs += vertices[pos].size();
		nodeOffsets[pos] = accumOffs;//store the accumulated offset in "offsets"
	}
	if (nodeOffsets[vertPositions.size()-1] != numVertices)
		std::cout << "Warning: ObjData::nodeOffsets[last] varies from numVertices. nodeOffsets[last]=" << nodeOffsets << \
		", numVertices=" << numVertices << std::endl;

}


GLint ObjData::getNumVertices(void)
{
	return numVertices;
}




GLfloat* ObjData::getVertPositions(void)
{
	GLfloat* vertexData = new GLfloat[numVertices*3];
	int i = 0;
	for (unsigned int pos = 0; pos < vertPositions.size(); pos++)
	{
		for (std::vector<VertexDataPointers>::iterator vert = vertices[pos].begin();
			vert != vertices[pos].end(); ++vert)
		{
			// Add positions
			vertexData[i++] = vertPositions[pos].x;
			vertexData[i++] = vertPositions[pos].y;
			vertexData[i++] = vertPositions[pos].z;
		}
	}

	if (i != (numVertices*3))
		std::cout << "WARNING: ObjData::getVertexdata() did not find the right amount of vertex positions! " << 
			"Was " << i << ", should've been " << numVertices*3 << std::endl;

	return vertexData;

}

GLfloat* ObjData::getVertNormals(void)
{
	GLfloat* vertexData = new GLfloat[numVertices*3];
	int i = 0;
	for (unsigned int pos = 0; pos < vertPositions.size(); pos++)
	{
		for (std::vector<VertexDataPointers>::iterator vert = vertices[pos].begin();
			vert != vertices[pos].end(); ++vert)
		{
			int norm = vert->normalIdx;
			// Add normals
			vertexData[i++] = vertNormals[norm].x;
			vertexData[i++] = vertNormals[norm].y;
			vertexData[i++] = vertNormals[norm].z;
		}
	}

	if (i != (numVertices*3))
		std::cout << "WARNING: ObjData::getVertexdata() did not find the right amount of vertex normals! " << 
			"Was " << i << ", should've been " << numVertices*3 << std::endl;

	return vertexData;

}

GLfloat* ObjData::getVertTexcoords(void)
{
	GLfloat* vertexData = new GLfloat[numVertices*3];
	int i = 0;
	for (unsigned int pos = 0; pos < vertPositions.size(); pos++)
	{
		for (std::vector<VertexDataPointers>::iterator vert = vertices[pos].begin();
			vert != vertices[pos].end(); ++vert)
		{
			int tex = vert->texIdx;
			// Add texture coords
			vertexData[i++] = vertTexture[tex].x;
			vertexData[i++] = vertTexture[tex].y;
			vertexData[i++] = vertTexture[tex].z;
		}
	}

	if (i != (numVertices*3))
		std::cout << "WARNING: ObjData::getVertexdata() did not find the right amount of texture coords! " << 
			"Was " << i << ", should've been " << numVertices*3 << std::endl;

	return vertexData;

}

GLfloat* ObjData::getVertColors(void)
{
	GLfloat* vertexData = new GLfloat[numVertices*3];
	int i = 0;
	for (unsigned int pos = 0; pos < vertPositions.size(); pos++)
	{
		for (std::vector<VertexDataPointers>::iterator vert = vertices[pos].begin();
			vert != vertices[pos].end(); ++vert)
		{
			int col = vert->colorIdx;
			// Add colors
			vertexData[i++] = vertColors[col].x;
			vertexData[i++] = vertColors[col].y;
			vertexData[i++] = vertColors[col].z;
		}
	}

	if (i != (numVertices*3))
		std::cout << "WARNING: ObjData::getVertColors() did not find the right number of vertex colors! " << 
			"Was " << i << ", should've been " << numVertices*3 << std::endl;

	return vertexData;

}

GLfloat* ObjData::getVertRadiuses(void)
{
	GLfloat* vertexData = new GLfloat[numVertices];
	int i = 0;
	for (unsigned int pos = 0; pos < vertPositions.size(); pos++)
	{
		for (std::vector<VertexDataPointers>::iterator vert = vertices[pos].begin();
			vert != vertices[pos].end(); ++vert)
		{
			int tex = vert->texIdx;
			// Add colors
			vertexData[i++] = vertRadius[tex];
		}
	}

	if (i != (numVertices*3))
		std::cout << "WARNING: ObjData::getVertRadiuses() did not find the right number of vertex radiuses! " << 
			"Was " << i << ", should've been " << numVertices << std::endl;

	return vertexData;

}

/********************
 * Returns a pointer to a (new) array of initial vertex displacement amounts. 
 * Normally this will be an array of 'numVertices' zeroes.
 */
GLfloat* ObjData::getVertDisplacements(void)
{
	GLfloat* vertexData = new GLfloat[numVertices];
	// hoping I can use memset on a float array, or else uncomment the for-loop instead....
	memset(vertexData, 0, numVertices*sizeof(GLfloat));
	//for (unsigned int node = 0; node < vertPositions.size(); node++)
	//	vertexData[pos] = 0.0f;
	return vertexData;
}


GLint ObjData::getNumFaces(void){
	return numFaces;
}

GLint ObjData::getNumVertsPrFace(void){
	return numVertsPrFace;
}


GLint* ObjData::getFaces(void){

	GLint* convFaces = new GLint[numFaces*numVertsPrFace];

	
	for (int vert = 0; vert < (numFaces*numVertsPrFace); vert++)
	{
		convFaces[vert] = faces[vert].no;
		// use the node offsets to calculate the new vertex-index in the face lookup table (pos 0 has no offset)
		if (faces[vert].pos > 0)
			convFaces[vert] += nodeOffsets[faces[vert].pos-1];
		
	}

	return convFaces;
}

GLint ObjData::getNumNodes(void){
	return vertPositions.size();
}

GLint* ObjData::getVertexNodeMap(void){
	// return a copy rather than the member variable, 
	// as this would have made a mess of the memory clean-up
	GLint* vnmcopy = new GLint[vertPositions.size()];
	memcpy(vnmcopy, nodeOffsets, vertPositions.size()*sizeof(GLint));
	return vnmcopy;
}

GLint ObjData::getMaxNbors(void){
	return numMaxNbors;
}

GLint* ObjData::getNborCounts(void)
{
	// return a copy rather than the member variable, 
	// as this would have made a mess of the memory clean-up
	GLint* nccopy = new GLint[vertPositions.size()];
	memcpy(nccopy, nborCounts, vertPositions.size()*sizeof(GLint));
	return nccopy;
}

#define NBORS(row,col) neighbourMatrix[((row)*numMaxNbors)+(col)]
/***********************
 * Returns a pointer to a neighbour matrix.
 * 
 */
GLint* ObjData::getNeighbourMatrix(void){
	// create a new neighbour matrix and fill it from the existing
	GLint* neighbourMatrix = new GLint[numVertices*numMaxNbors];
	int i = 0;
	for (unsigned int node = 0; node < vertPositions.size(); node++)
	{
		for (int nbor = 0; nbor < nborCounts[node]; nbor++)
		{
			NBORS(node,nbor) = neighbours[node][nbor];
		}
		// mark any remaining cells as empty, just in case...
		for (int nbor = nborCounts[node]; nbor < numMaxNbors; nbor++)
		{
			NBORS(node,nbor) = -1;
		}
	}
	return neighbourMatrix;
}

#define DISTS(row,col) distMatrix[((row)*numMaxNbors)+(col)]

GLfloat* ObjData::getDistanceMatrix(void){
	// create a new distances matrix and fill it from the existing
	GLfloat* distMatrix = new GLfloat[numVertices*numMaxNbors];
	//int i = 0;
	for (unsigned int node = 0; node < vertPositions.size(); node++)
	{
		for (int nbor = 0; nbor < nborCounts[node]; nbor++)
		{
			DISTS(node,nbor) = distances[node][nbor];
		}
		// mark any remaining cells as empty, just in case...
		for (int nbor = nborCounts[node]; nbor < numMaxNbors; nbor++)
		{
			DISTS(node,nbor) = -1.0;
		}
	}
	return distMatrix;
}



#define LINKS(row,col,coord) linkMatrix[((((row)*(numMaxNbors))+(col))*3)+(coord)]
GLfloat* ObjData::getLinkdirMatrix()
{
	GLfloat* linkMatrix = new GLfloat[numVertices*numMaxNbors*3];
	for (unsigned int node = 0; node < vertPositions.size(); node++)
	{
		for (int nbor = 0; nbor < nborCounts[node]; nbor++)
		{			
			Vec3f dir = getLinkDir(node,nbor);
			LINKS(node,nbor,0) = dir.x;
			LINKS(node,nbor,1) = dir.y;
			LINKS(node,nbor,2) = dir.z;
		}
		// mark any remaining cells as empty, just in case...
		for (int nbor = nborCounts[node]; nbor < numMaxNbors; nbor++)
		{
			LINKS(node,nbor,0) = -1.0;
			LINKS(node,nbor,1) = -1.0;
			LINKS(node,nbor,2) = -1.0;
		}
	}
	return linkMatrix;

}


GLfloat* ObjData::getNodeFuels(void){
	// create a new distances matrix and fill it from the existing
	GLfloat* nodeFuels = new GLfloat[vertPositions.size()];
	int i = 0;
	for (unsigned int node = 0; node < vertPositions.size(); node++)
	{
		int vertTexIdx = vertices[node][0].texIdx; // All vertices at one position should have the same fuel value (radius)
		nodeFuels[node] = vertRadius[vertTexIdx];
	}
	return nodeFuels;
}

