#include "StdAfx.h"
#include "RenderObject.h"




RenderObject::RenderObject(void)
{
	// Load the identity matrix into the model, view, and projection matrices.
    loadIdMatrix4x4f(m_modelMatrix);
}

RenderObject::RenderObject(GLfloat* vertexPositions, GLfloat* vertexNormals, GLfloat* vertexColors, 
						   GLfloat* vertexTexcoords, GLfloat* vertexDispl,
							 GLint* facelookuptable, GLint numVertices, GLint numDrawVerts, 
							 const char* vertShaderFilename, const char* fragShaderFilename, const char* textureFilename)
{
	// Load the identity matrix into the model matrix.
    loadIdMatrix4x4f(m_modelMatrix);

	this->numDrawVerts = numDrawVerts;
	if(!loadShaderProgram(vertShaderFilename, fragShaderFilename))
	{
		pausePressEnter();
		exit(1);
	}
	makeVertexBuffers(vertexPositions, vertexNormals, vertexColors, vertexTexcoords, vertexDispl, 
		facelookuptable, numVertices);
	initTexture(textureFilename);
	delete[] vertexPositions;
	delete[] vertexNormals;
	delete[] vertexColors;
	delete[] vertexTexcoords;
	delete[] facelookuptable;

}

RenderObject::~RenderObject(void)
{

	if (vertexArrayObject)
    {
        glDeleteVertexArrays(1, &vertexArrayObject);
        vertexArrayObject = 0;
    }

    if (vertexBuffersObject)
    {
        glDeleteBuffers(NUM_VERTEX_BUFFERS, vertexBuffersObject);
		for (int i = 0; i < NUM_VERTEX_BUFFERS; i++)
			vertexBuffersObject[i] = 0;
    }

    if (indexBufferObject)
    {
        glDeleteBuffers(1, &indexBufferObject);
        indexBufferObject = 0;
    }

	if (texObject)
	{
		glDeleteTextures(1, &texObject);
		texObject = 0;
	}


    glUseProgram(0);

    if (shaderProgram)
    {
        glDeleteProgram(shaderProgram);
        shaderProgram = 0;
    }

}

void RenderObject::reset(void)
{
	// Load the identity matrix into the model matrix.
    loadIdMatrix4x4f(m_modelMatrix);

}

void RenderObject::initTexture(const char* filename)
{
	const int mipLevel = 0;
	const int border = 0;
	const int internFormat = GL_RGBA;
	const int format = GL_RGBA;
	const int type = GL_UNSIGNED_BYTE;
	const int texUnit = 0;

	unsigned long width, height;
  
	//load and decode texture
	std::vector<unsigned char> buffer, image;
	loadFile(buffer, filename);
	int error = decodePNG(image, width, height, buffer.empty() ? 0 : &buffer[0], (unsigned long)buffer.size());
	if(error != 0) 
		std::cout << "error: " << error << std::endl;
	// copy texture to a normal array. 
	GLubyte* data = new GLubyte[width*height*4];
	for (int i = 0; i < width*height*4; i++)
	{
		data[i] = image[i];
	}

	//the pixels are now in the vector "image", use it as texture, draw it, ...
	if(image.size() > 4) 
		std::cout << "width: " << width << " height: " << height << 
			" first pixel: " << //std::hex << 
			int(image[0]) << ", " << int(image[1]) << ", " << 
			int(image[2]) << ", " << int(image[3]) << std::endl;
	int testidx = 10;
	std::cout << "10th pixel of data: " << //std::hex << 
		int(data[testidx*4+0]) << ", " << int(data[testidx*4+1]) << ", " << 
		int(data[testidx*4+2]) << ", " << int(data[testidx*4+3]) << std::endl;
	//pausePressEnter();
	
	//glEnable(GL_TEXTURE_2D);
	//checkError("RenderObject::initTexture, initial gen and bind.");
	glActiveTexture(GL_TEXTURE0);// + texUnit);
	glGenTextures(1, &texObject);
	glBindTexture(GL_TEXTURE_2D, texObject);
	checkError("RenderObject::initTexture, initial gen and bind.");
	// set parameters here:
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexImage2D(GL_TEXTURE_2D, mipLevel, internFormat, width, height, border, format, type, data);
	checkError("RenderObject::initTexture, glTexImage().");

	GLint tmpIdx = glGetUniformLocation(shaderProgram, "BarkTexture");
	checkError("RenderObject::initTexture, uniform get location.");
	glUseProgram(shaderProgram); // need to bind the right program first
	glUniform1i(tmpIdx, texUnit);
	checkError("RenderObject::initTexture, uniform set.");
	std::cout << "BarkTexture index is " << tmpIdx << std::endl;
	//pausePressEnter();

	
	glBindTexture(GL_TEXTURE_2D, 0); //unbind
    glUseProgram(0);
	checkError("RenderObject::initTexture, unbind texture.");

	delete[] data;


}


void RenderObject::makeVertexBuffers(GLfloat* vertPos, GLfloat* vertNorm, GLfloat* vertColors, GLfloat* vertTexture, 
									GLfloat* vertDispl, GLint* facelookuptable, GLint numVertices)
{

	std::cout << "DEBUG: Printing sample of incoming vertex data and facelookups for RenderObject::makeVertexBuffer:" <<std::endl;
	for (int i = 0; i < 3; i++)
	{
		std::cout << "  Vertex " << i << " has the following data..." << std::endl <<
			"   Position: (" << vertPos[3*i] << ", " << vertPos[3*i+1] << ", " << vertPos[3*i+2] << ")" << std::endl <<\
			"   Normal: (" << vertNorm[3*i] << ", " << vertNorm[3*i+1] << ", " << vertNorm[3*i+2] << ")" << std::endl <<\
			"   Color: (" << vertColors[3*i] << ", " << vertColors[3*i+1] << ", " << vertColors[3*i+2] << ")" << std::endl;
			
	}
	std::cout << "  The facelookuptable has the following first 30 entries:" << std::endl << "  ";
	for (int i = 0; i < 30; i++)
	{
		std::cout << facelookuptable[i] << ", ";
	}
	std::cout << "." << std::endl;

	GLfloat* vertGlow = new GLfloat[numVertices*4];
	memset(vertGlow, 0, numVertices*4*sizeof(GLfloat));
	

	glGenVertexArrays(1, &vertexArrayObject);
    glBindVertexArray(vertexArrayObject);
	checkError("RenderObject::makeVertexBuffers, vertex array generation.");

    // Create the Index Buffer
    glGenBuffers(1, &indexBufferObject);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, indexBufferObject);
    //glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(GLuint) * elementCounts.numFacesIndices, tempIndices, GL_STATIC_DRAW);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(GLuint) * numDrawVerts, facelookuptable, GL_STATIC_DRAW);
	checkError("RenderObject::makeVertexBuffers, index buffer creation and filling.");

    // Create the Vertex Buffers
    // Buffer names:
    glGenBuffers(NUM_VERTEX_BUFFERS, vertexBuffersObject);
	// Position buffer:
	glBindBuffer(GL_ARRAY_BUFFER, vertexBuffersObject[POSITION_BUFFER]);
    glBufferData(GL_ARRAY_BUFFER, sizeof(Vec3f) * numVertices, vertPos, GL_STATIC_DRAW);
	checkError("RenderObject::makeVertexBuffers, position buffer creation and filling.");
	GLint tmpIdx = glGetAttribLocation(shaderProgram, "in_Position");
	checkError("RenderObject::makeVertexBuffers, position attrib get location");
	glVertexAttribPointer(tmpIdx , 3, GL_FLOAT, GL_FALSE, 0, BUFFER_OFFSET(0));
	checkError("RenderObject::makeVertexBuffers, position attrib pointer binding");
	GLint tmp = 0;
    glGetIntegerv(GL_MAX_VERTEX_ATTRIBS, &tmp);
	std::cout << "GL_MAX_VERTEX_ATTRIBS " << tmp << ", index: " << tmpIdx << std::endl;
	//pausePressEnter();
	//exit(1);
    glEnableVertexAttribArray(glGetAttribLocation(shaderProgram, "in_Position"));
	checkError("RenderObject::makeVertexBuffers, position attrib enabling");
	// Normal buffer
	glBindBuffer(GL_ARRAY_BUFFER, vertexBuffersObject[NORMAL_BUFFER]);
    glBufferData(GL_ARRAY_BUFFER, sizeof(Vec3f) * numVertices, vertNorm, GL_STATIC_DRAW);
	checkError("RenderObject::makeVertexBuffers, normal buffer creation and filling.");
    glVertexAttribPointer(glGetAttribLocation(shaderProgram, "in_Normal"), 3, GL_FLOAT, GL_FALSE, 0, BUFFER_OFFSET(0));
    glEnableVertexAttribArray(glGetAttribLocation(shaderProgram, "in_Normal"));
	checkError("RenderObject::makeVertexBuffers, normal attrib pointer binding");
	// Texcoords buffer
	glBindBuffer(GL_ARRAY_BUFFER, vertexBuffersObject[TEX_BUFFER]);
    glBufferData(GL_ARRAY_BUFFER, sizeof(Vec3f) * numVertices, vertTexture, GL_STATIC_DRAW);
	checkError("RenderObject::makeVertexBuffers, texCoord buffer creation and filling.");
    glVertexAttribPointer(glGetAttribLocation(shaderProgram, "in_TexCoord"), 3, GL_FLOAT, GL_FALSE, 0, BUFFER_OFFSET(0));
    glEnableVertexAttribArray(glGetAttribLocation(shaderProgram, "in_TexCoord"));
	checkError("RenderObject::makeVertexBuffers, texCoord attrib pointer binding");
	// Color buffer:
	/* Dont need colors when I have texture.
	glBindBuffer(GL_ARRAY_BUFFER, vertexBuffersObject[COLOR_BUFFER]);
    glBufferData(GL_ARRAY_BUFFER, sizeof(Vec3f) * numVertices, vertColors, GL_DYNAMIC_DRAW);
	checkError("RenderObject::makeVertexBuffers, color buffer creation and filling.");
    glVertexAttribPointer(glGetAttribLocation(shaderProgram, "in_Color"), 3, GL_FLOAT, GL_FALSE, 0, BUFFER_OFFSET(0));
    glEnableVertexAttribArray(glGetAttribLocation(shaderProgram, "in_Color"));
	checkError("RenderObject::makeVertexBuffers, color attrib pointer binding");
	*/
	// Glow buffer:
	glBindBuffer(GL_ARRAY_BUFFER, vertexBuffersObject[GLOW_BUFFER]);
    glBufferData(GL_ARRAY_BUFFER, 4*sizeof(GLfloat)*numVertices, vertGlow, GL_DYNAMIC_DRAW);
	checkError("RenderObject::makeVertexBuffers, glow buffer creation and filling.");
    glVertexAttribPointer(glGetAttribLocation(shaderProgram, "in_Glow"), 4, GL_FLOAT, GL_FALSE, 0, BUFFER_OFFSET(0));
    glEnableVertexAttribArray(glGetAttribLocation(shaderProgram, "in_Glow"));
	checkError("RenderObject::makeVertexBuffers, color attrib pointer binding");
	// Displacement buffer:
	glBindBuffer(GL_ARRAY_BUFFER, vertexBuffersObject[DISPL_BUFFER]);
    glBufferData(GL_ARRAY_BUFFER, sizeof(GLfloat) * numVertices, vertDispl, GL_DYNAMIC_DRAW);
	checkError("RenderObject::makeVertexBuffers, displ buffer creation and filling.");
    glVertexAttribPointer(glGetAttribLocation(shaderProgram, "in_Displ"), 1, GL_FLOAT, GL_FALSE, 0, BUFFER_OFFSET(0));
    glEnableVertexAttribArray(glGetAttribLocation(shaderProgram, "in_Displ"));
	checkError("RenderObject::makeVertexBuffers, displ attrib pointer binding");

	delete[] vertGlow;

}



void RenderObject::draw(RenderContext* renderContextp)
{
	checkError("draw -1");
    glUseProgram(shaderProgram);
	checkError("draw 0");
	
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, texObject);

	GLfloat tmp[3][3];

	// make normal-matrix
	memcpy(&tmp[0][0], &m_modelMatrix[0], 3*sizeof(GLfloat));
	memcpy(&tmp[1][0], &m_modelMatrix[4], 3*sizeof(GLfloat));
	memcpy(&tmp[2][0], &m_modelMatrix[8], 3*sizeof(GLfloat));
	if (!inverse(tmp,m_normalMatrix))
		std::cerr << "Warning: normal matrix not constructed because inversing erred.";
	const GLfloat* nMatp = &m_normalMatrix[0][0];


    glUniformMatrix4fv(glGetUniformLocation(shaderProgram, "ProjectionMatrix"), 1, GL_FALSE, renderContextp->m_projectionMatrix);
	checkError("draw 1");
    glUniformMatrix4fv(glGetUniformLocation(shaderProgram, "ViewMatrix"), 1, GL_FALSE, renderContextp->m_viewMatrix);
	checkError("draw 2");
    glUniformMatrix4fv(glGetUniformLocation(shaderProgram, "ModelMatrix"), 1, GL_FALSE, m_modelMatrix);
	checkError("draw 3");
	glUniformMatrix3fv(glGetUniformLocation(shaderProgram, "NormalMatrix"), 1, GL_TRUE, nMatp); //load and transpose
	checkError("draw 3");

	//set lightposition
	glUniform3fv(glGetUniformLocation(shaderProgram, "LightPosition"), 1, renderContextp->lightPosition);
	checkError("draw lightpos");

    glBindVertexArray(vertexArrayObject);
	checkError("draw 4");
	glDrawElements(GL_TRIANGLES, numDrawVerts, GL_UNSIGNED_INT, BUFFER_OFFSET(0));
	checkError("draw 5");
	
    glBindVertexArray(0);

    glBindTexture(GL_TEXTURE_2D, 0);
    glUseProgram(0);

	//DEBUG: print matrices
	/*printMatrix3("NormalMatrix", m_normalMatrix);
	printMatrix16("ProjectionMatrix", renderContextp->m_projectionMatrix);
	printMatrix16("ViewMatrix", renderContextp->m_viewMatrix);
	printMatrix16("ModelMatrix", m_modelMatrix);
	*/

}





/* Perform translation operations on a matrix */
// Stolen from some tutorial. Don't remember which...
void RenderObject::translate(GLfloat dx, GLfloat dy, GLfloat dz)
{
    GLfloat newmatrix[16] = IDENTITY_MATRIX4;

    newmatrix[12] = dx;
    newmatrix[13] = dy;
    newmatrix[14] = dz;

    multiply4x4(m_modelMatrix, newmatrix);
}


/* Rotate a matrix by an angle on a X, Y, or Z axis */
// Stolen from some tutorial. Don't remember which...
void RenderObject::rotate(GLfloat angle, AXIS axis)
{
    const GLfloat d2r = PI_OVER_180;
    const int cos1[3] = { 5, 0, 0 };
    const int cos2[3] = { 10, 10, 5 };
    const int sin1[3] = { 6, 2, 1 };
    const int sin2[3] = { 9, 8, 4 };
    GLfloat newmatrix[16] = IDENTITY_MATRIX4;

    newmatrix[cos1[axis]] = cos(d2r * angle);
    newmatrix[sin1[axis]] = -sin(d2r * angle);
    newmatrix[sin2[axis]] = -newmatrix[sin1[axis]];
    newmatrix[cos2[axis]] = newmatrix[cos1[axis]];

    multiply4x4(m_modelMatrix, newmatrix);
}


/* Scale a matrix in the X, Y, or Z axis */
void RenderObject::scale(GLfloat sx, GLfloat sy, GLfloat sz)
{
    GLfloat newmatrix[16] = IDENTITY_MATRIX4;

    newmatrix[0] = sx;
    newmatrix[5] = sy;
    newmatrix[10] = sz;

    multiply4x4(m_modelMatrix, newmatrix);
}




std::string shaderInfoLog(GLuint shader)// const
{
    GLsizei infoLogSize = 0;
    std::string infoLog;

    glGetShaderiv(shader, GL_INFO_LOG_LENGTH, &infoLogSize);
    infoLog.resize(infoLogSize);
    glGetShaderInfoLog(shader, infoLogSize, &infoLogSize, &infoLog[0]);

    return infoLog;
}

std::string programInfoLog(GLuint program)// const
{
    GLsizei infoLogSize = 0;
    std::string infoLog;

    glGetProgramiv(program, GL_INFO_LOG_LENGTH, &infoLogSize);
    infoLog.resize(infoLogSize);
    glGetProgramInfoLog(program, infoLogSize, &infoLogSize, &infoLog[0]);

    return infoLog;
}

#define MAX_SHADER_SOURCE_LENGTH 10000
/* Takes the vertex-shader- and fragment-shader-source as null-terminated strings */
// Stolen from some tutorial. Don't remember which...
GLboolean RenderObject::loadShaderProgram(const GLchar* vertexSourceFile, const GLchar* fragmentSourceFile)
{
	
	std::string m_shaderInfoLog;
    std::string m_programInfoLog;
    shaderProgram = 0;
    GLint compiled = 0;
    GLint linked = 0;
    GLuint vertexShader = 0;
    GLuint fragmentShader = 0;
	GLchar* tmpCharp = NULL;
	//GLchar* fragmentSource = NULL;
	//const char * code = m_code;

	if((tmpCharp = readFile(vertexSourceFile, MAX_SHADER_SOURCE_LENGTH)) == NULL)
	{
		std::cerr << "Could  not open vertex shader source file '" << vertexSourceFile << "'." << std::endl;
		return GL_FALSE;
	} 
	const GLchar* vertexSource = tmpCharp;
	
	if((tmpCharp = readFile(fragmentSourceFile, MAX_SHADER_SOURCE_LENGTH)) == NULL)
	{
		std::cerr << "Could  not open vertex shader source file '" << fragmentSourceFile << "'." << std::endl;
		return GL_FALSE;
	} 
	const GLchar* fragmentSource = tmpCharp;
	

	vertexShader = glCreateShader(GL_VERTEX_SHADER);
	fragmentShader = glCreateShader(GL_FRAGMENT_SHADER);

	glShaderSource(vertexShader, 1, (const GLchar**) &vertexSource, 0);
	glShaderSource(fragmentShader, 1, (const GLchar**) &fragmentSource, 0);


	/* GL3Application::loadShaderProgram() */
	//vertexShader = loadVertexShader(pszVertexShaderFilename);
    glCompileShader(vertexShader);
    glGetShaderiv(vertexShader, GL_COMPILE_STATUS, &compiled);

	delete [] fragmentSource;
	delete [] vertexSource;

    if (!compiled)
    {
        m_shaderInfoLog = shaderInfoLog(vertexShader);
        glDeleteShader(vertexShader);

		std::cerr << "Error compiling GLSL vertex shader: '";
        std::cerr << vertexSourceFile;
        std::cerr << "'" << std::endl << std::endl;
        std::cerr << "Shader info log:" << std::endl;
        std::cerr << m_shaderInfoLog;
		return GL_FALSE;
    }
    else
    {
        //fragmentShader = loadFragmentShader(pszFragmentShaderFilename);
        glCompileShader(fragmentShader);
        glGetShaderiv(fragmentShader, GL_COMPILE_STATUS, &compiled);

        if (!compiled)
        {
            m_shaderInfoLog = shaderInfoLog(fragmentShader);
            glDeleteShader(vertexShader);
            glDeleteShader(fragmentShader);


            std::cerr << "Error compiling GLSL fragment shader: '";
            std::cerr << fragmentSourceFile;
            std::cerr << "'" << std::endl << std::endl;
            std::cerr << "Shader info log:" << std::endl;
            std::cerr << m_shaderInfoLog;

			return GL_FALSE;
        }
        else
        {
            shaderProgram = glCreateProgram();
            glAttachShader(shaderProgram, vertexShader);
            glAttachShader(shaderProgram, fragmentShader);
            glLinkProgram(shaderProgram);
            glGetProgramiv(shaderProgram, GL_LINK_STATUS, &linked);

            // Mark the two attached shaders for deletion. These two shaders
            // aren't deleted right now because both are already attached to a
            // shader program. When the shader program is deleted these two
            // shaders will be automatically detached and deleted.

            glDeleteShader(vertexShader);
            glDeleteShader(fragmentShader);

            if (!linked)
            {
                m_programInfoLog = programInfoLog(shaderProgram);
                glDeleteProgram(shaderProgram);

                std::cerr << "Error linking GLSL shaders into a shader program." << std::endl;
                std::cerr << "GLSL vertex shader: '" << vertexSourceFile;
                std::cerr << "'" << std::endl;
                std::cerr << "GLSL fragment shader: '" << fragmentSourceFile;
                std::cerr << "'" << std::endl << std::endl;
                std::cerr << "Program info log:" << std::endl;
                std::cerr << m_programInfoLog;

				return GL_FALSE;
            }
        }
    }


    return GL_TRUE;
}




