#include "StdAfx.h"
#include "utils.h"

/* A simple function that will read a file into a dynamically allocated char pointer buffer */
GLchar* readFile(const GLchar* filename, long maxlength)
{
	std::ifstream in;
    long length;
    char* buf;
	char* tempbuf = new char[maxlength];
	in.open(filename);
	if(!in) {
		std::cout << "Cannot open file: " << filename << ".\n";
		return NULL;
	}
	in.read(tempbuf, maxlength);
	
	if(!in.eof())
	{
		std::cout << "File was longer than maxlength (" << maxlength << ") - didn't reach eof." << std::endl;
	}

	length = in.gcount();

	buf = new char[length+1]; //+1 makes room for the terminating null
	buf[length] = NULL;

	memcpy(buf, tempbuf, length * sizeof(char));

	delete [] tempbuf;

	in.close();

    return buf; /* Return the buffer */
}

/* Another one using vector. From the example for picopng by Lode Vandevenne. */
void loadFile(std::vector<unsigned char>& buffer, const std::string& filename) //designed for loading files from hard disk in an std::vector
{
  std::ifstream file(filename.c_str(), std::ios::in|std::ios::binary|std::ios::ate);

  //get filesize
  std::streamsize size = 0;
  if(file.seekg(0, std::ios::end).good()) size = file.tellg();
  if(file.seekg(0, std::ios::beg).good()) size -= file.tellg();

  //read contents of the file into the vector
  if(size > 0)
  {
    buffer.resize((size_t)size);
    file.read((char*)(&buffer[0]), size);
  }
  else buffer.clear();
}

void pausePressEnter()
{
	system("pause");
}

void printVec3f(Vec3f* vec)
{
	std::cout << "(" << vec->x << ", " << vec->y << ", " << vec->z << ")";
}

/* Multiply 4x4 matrix m1 by 4x4 matrix m2 and store the result in m1 */
void multiply4x4(GLfloat *m1, GLfloat *m2)
{
    GLfloat temp[16];

    int x,y;

    for (x=0; x < 4; x++)
    {
        for(y=0; y < 4; y++)
        {
            temp[y + (x*4)] = (m1[x*4] * m2[y]) +
                              (m1[(x*4)+1] * m2[y+4]) +
                              (m1[(x*4)+2] * m2[y+8]) +
                              (m1[(x*4)+3] * m2[y+12]);
        }
    }

    memcpy(m1, temp, sizeof(GLfloat) << 4);
}

/* glmAbs: returns the absolute value of a float */
static GLfloat
glmAbs(GLfloat f)
{
    if (f < 0)
        return -f;
    return f;
}

/* glmEqual: compares two vectors and returns GL_TRUE if they are
 * equal (within a certain threshold) or GL_FALSE if not. An epsilon
 * that works fairly well is 0.000001.
 *
 * u - array of 3 GLfloats (GLfloat u[3])
 * v - array of 3 GLfloats (GLfloat v[3]) 
 */
GLboolean vec3Equal(Vec3f u, Vec3f v, GLfloat epsilon)
{
	if (glmAbs(u.x - v.x) < epsilon &&
		glmAbs(u.y - v.y) < epsilon &&
		glmAbs(u.z - v.z) < epsilon) 
    {
        return GL_TRUE;
    }
    return GL_FALSE;
}

GLfloat determinantOfMinor( int          theRowHeightY,
                           int          theColumnWidthX,
                           const GLfloat theMatrix [/*Y=*/3] [/*X=*/3] )
{
  int x1 = theColumnWidthX == 0 ? 1 : 0;  /* always either 0 or 1 */
  int x2 = theColumnWidthX == 2 ? 1 : 2;  /* always either 1 or 2 */
  int y1 = theRowHeightY   == 0 ? 1 : 0;  /* always either 0 or 1 */
  int y2 = theRowHeightY   == 2 ? 1 : 2;  /* always either 1 or 2 */

  return ( theMatrix [y1] [x1]  *  theMatrix [y2] [x2] )
      -  ( theMatrix [y1] [x2]  *  theMatrix [y2] [x1] );
}

GLfloat determinant( const GLfloat theMatrix [/*Y=*/3] [/*X=*/3] )
{
  return ( theMatrix [0] [0]  *  determinantOfMinor( 0, 0, theMatrix ) )
      -  ( theMatrix [0] [1]  *  determinantOfMinor( 0, 1, theMatrix ) )
      +  ( theMatrix [0] [2]  *  determinantOfMinor( 0, 2, theMatrix ) );
}


// Courtesy mrree at stackoverflow
GLboolean inverse( const GLfloat theMatrix [/*Y=*/3] [/*X=*/3],
                    GLfloat theOutput [/*Y=*/3] [/*X=*/3] )
{
  GLfloat det = determinant( theMatrix );

    /* Arbitrary for now.  This should be something nicer... */
  if ( glmAbs(det) < 1e-2 )
  {
    memset( theOutput, 0, sizeof theOutput );
    return false;
  }

  GLfloat oneOverDeterminant = 1.0f / det;

  for (   int y = 0;  y < 3;  y ++ )
    for ( int x = 0;  x < 3;  x ++   )
    {
        /* Rule is inverse = 1/det * minor of the TRANSPOSE matrix.  *
         * Note (y,x) becomes (x,y) INTENTIONALLY here!              */
      theOutput [y] [x]
        = determinantOfMinor( x, y, theMatrix ) * oneOverDeterminant;

        /* (y0,x1)  (y1,x0)  (y1,x2)  and (y2,x1)  all need to be negated. */
      if( 1 == ((x + y) % 2) )
        theOutput [y] [x] = - theOutput [y] [x];
    }

  return true;
}

void transpose(GLfloat a[3][3], GLfloat b[3][3])
{
   int i, j;

   for(i=0;i<3;i++)
      for (j=0; j<3;j++)
      {
         b[j][i]=a[i][j];
      }

}



int checkError(const char *location)//, const char *file, int line)
{
	//const char *file = GLcall;//"RenderObject.cpp";
	GLenum glErr;
	int retCode = 0;
	while ( (glErr=glGetError()) != GL_NO_ERROR)
	{
		switch(glErr)
		{
			case GL_INVALID_VALUE:
			fprintf (stderr, "GL_INVALID_VALUE error in function %s.\n", location);
			break;
			case GL_INVALID_ENUM:
			fprintf (stderr, "GL_INVALID_ENUM error in function %s.\n", location);
			break;
			case GL_INVALID_OPERATION:
			fprintf (stderr, "GL_INVALID_OPERATION error in function %s.\n", location);
			break;
			case GL_STACK_OVERFLOW:
			fprintf (stderr, "GL_STACK_OVERFLOW error in function %s.\n", location);
			break;
			case GL_STACK_UNDERFLOW:
			fprintf (stderr, "GL_STACK_UNDERFLOW error in function %s.\n", location);
			break;
			case GL_OUT_OF_MEMORY:
			fprintf (stderr, "GL_OUT_OF_MEMORY error in function %s.\n", location);
			break;
			default:
			fprintf (stderr, "UNKNOWN ERROR in function %s.\n", location);
		}
	}
	return retCode;
}

void printMatrix3(const char* name, const GLfloat m[3][3])
{
	std::cout << "3x3 matrix '" << name << "':" << std::endl;
	for (int i = 0; i < 3; i++)
	{
		for (int j = 0; j < 3; j++)
		{
			std::cout << " | " << m[i][j];
		}
		std::cout << std::endl;
	}
	std::cout << std::endl;
}

void printMatrix16(const char* name, const GLfloat m[16])
{
	std::cout << "4x4 matrix '" << name << "':" << std::endl;
	for (int i = 0; i < 4; i++)
	{
		for (int j = 0; j < 4; j++)
		{
			std::cout << " | " << m[i*4+j];
		}
		std::cout << std::endl;
	}
	std::cout << std::endl;
}



